"""Formulaire PI."""
from __future__ import annotations

from labster.domain.services.constants import get_constant
from labster.newforms.base import HTML, Boolean2Field, BooleanField, \
    DateField, FieldSet, Form, IntegerField, ListeAutresDeclarations, \
    ListeContrats, ListeLicencesExistantes, Select2Field, StringField, \
    TextAreaField
from labster.newforms.common import laboratoire
from labster.newforms.pi_common import exploitation, porteur_de_projet

_CLASSES_PRODUIT_STR = """
10000 LOGICIEL SYSTÈME
- 10100 Systèmes d’exploitation
- 10200 Transmission de données
- 10300 Bases de données (SGBD)
- 10400 Langage de programmation
- 10500 Langage adapté à l’utilisateur final
- 10600 Assistance pour le développement
- 10700 Gestion d’exploitation de système
- 10800 Utilitaire
20000 LOGICIEL UNIVERSEL
- 20100 Planification / Gestion
- 20200 Comptabilité
- 20300 Personnel / Salaires
- 20400 Ventes / Inventaire
- 20500 Production
- 20600 Conception / étude / projet
- 20700 Prévision / Statistique / Analyse
- 20800 Bureautique
- 20900 Ingénierie d’information
- 21000 Traitement d’images
- 21100 EAO
30000 LOGICIEL SPECIALISE
- 30100 Agriculture
- 30200 Eaux et Forêts
- 30300 Pêche
- 30400 Exploitation minière
- 30500 Construction
- 30600 Production
- 30605 Alimentation
- 30610 Textile et habillement
- 30615 Bois, pâte et papier
- 30620 Publication et impression
- 30625 Chimie et industries dérivées
- 30630 Produits en pierre, argile et verre
- 30635 Produits métalliques
- 30640 Machines et Matériels
- 30645 Machines électriques
- 30650 Matériels de transport
- 30700 Fourniture électricité/gaz/chaleur/eau
- 30705 Electricité
- 30710 Gaz
- 30715 Fourniture d’eau
- 30800 Transport / Communication
- 30805 Transport
- 30810 Communication
- 30900 Ventes / Restaurants et débits de boissons
- 30905 Vente en gros
- 30910 Vente au détail
- 30915 Restaurants et débits de boissons
- 31000 Activités financières et d’assurances
- 31005 Activités bancaires et fiduciaires
- 31010 Courtage financier
- 31015 Assurances
- 31100 Immobilier
- 31200 Services
- 31205 Location
- 31210 Hôtels et auberges
- 31215 Radiodiffusion et publicité
- 31220 Services d’information
- 31225 Services divers
- 31230 Médecine, santé et sanitaire
- 31235 Education et recherche scientifique
- 31300 Services publics
- 31400 Loisirs et vie familiale
- 31500 Autres
40000 MULTIMEDIA / BASES DE DONNEES
- 40100 Reproduction numérique d’une œuvre 2D (peinture, photo, texte,...)
- 40200 Reproduction numérique d’œuvre 3D
- 40300 Reproduction numérique d’image animée
- 40400 Reproduction numérique d’un son
- 41000 CREATION NUMERIQUE
- 41100 Création numérique 2D
- 41200 Création numérique 3D
- 41300 Création d’une image animée
- 41400 Création numérique d’un son
- 41500 Création d’une photo numérique
- 42000 SITE WEB
"""

CLASSES_PRODUIT = _CLASSES_PRODUIT_STR.strip().split("\n")

#
# Fieldsets
#
description_oeuvre = FieldSet(
    "description_oeuvre",
    "Description de l'œuvre",
    [
        StringField("intitule", "Intitulé de l’œuvre", required=True),
        StringField("acronyme", "Acronyme", required=True),
        StringField("numero_version", "Numéro de version de l'œuvre", required=True),
        DateField("date_version", "Date de la version", required=True),
        IntegerField("nbre_lignes_code", "Nombre de lignes de code"),
        StringField("nationalite", "Nationalité de l'œuvre", required=True),
    ],
)

CHOICES1 = [
    "Logiciel (code source)",
    "Logiciel (code objet)",
    "Base de données",
    "Autre",
]

CHOICES_ORIGINALITE = [
    "logiciel premier (logiciel qui a été intégralement développé, sans utilisation de code préexistant)",
    "logiciel composé (logiciel qui incorpore tout ou partie d’un logiciel préexistant)",
    "logiciel dérivé (adaptation d’un logiciel préexistant, par exemple une version d’un programme pour un client spécifique ou portage d’un logiciel d’un environnement à un autre",
]

type_oeuvre = FieldSet(
    "type_oeuvre",
    "Type de l'œuvre",
    [
        Select2Field("type_oeuvre", "L'œuvre est", choices=CHOICES1),
        StringField("precision_type_oeuvre", "Précisez", required=True),
        Select2Field("originalite", "S'agit-il d'un", choices=CHOICES_ORIGINALITE),
        StringField(
            "logiciel_compose_details",
            "Indiquez les composants intégrés",
            required=True,
        ),
        StringField(
            "logiciel_derive_details",
            "Indiquez de quel logiciel il découle",
            required=True,
        ),
        StringField("domaine_application", "Domaine d’application", required=True),
        Select2Field(
            "classe_produit",
            "Classe de produit",
            choices=CLASSES_PRODUIT,
            required=True,
        ),
    ],
)

description_technique = FieldSet(
    "description_technique",
    "Description technique",
    [
        StringField(
            "langage_de_programmation", "Langage de programmation", required=True
        ),
        StringField(
            "outils_de_developpement", "Outils de développement", required=True
        ),
        BooleanField("cle_logique", "Clé logique/moyen cryptographique", required=True),
        HTML(
            """Joignez en PJ un schéma de l’architecture du logiciel, en indiquant les
            zones fonctionnelles, les différents modules et les modules
            facultatifs au bon fonctionnement du logiciel le cas échéant
            (avec l'onglet <b>Pièces à joindre</b>, une fois que vous avez enregistré le formulaire).
            """
        ),
        TextAreaField(
            "environnement",
            "Décrivez l’environnement du logiciel et ses interactions éventuelles avec d’autres logiciels/matériels (forge, GitHub, intégration continue, etc.).",
            required=True,
        ),
    ],
)

divulgation = FieldSet(
    "divulgation",
    "Divulgation",
    [
        BooleanField(
            "code_publie",
            "Le code source/objet a-t-il été publié, communiqué, mis en ligne ou fait l’objet d’un dépôt&nbsp;?",
            required=True,
        ),
        TextAreaField(
            "precision_code_publie",
            "Veuillez préciser",
            note="Joindre une copie des publications (avec l'onglet <b>Pièces à joindre</b>, une fois que vous aurez enregistré le formulaire).",
            required=True,
        ),
        BooleanField(
            "communication_prevue",
            "Est-ce qu’une communication relative à l’œuvre est prévue ?",
            required=True,
        ),
        TextAreaField(
            "precision_communication_prevue",
            "Veuillez préciser le type de communication et la date prévue",
            note="Joindre une copie du projet de communication (avec l'onglet <b>Pièces à joindre</b>, une fois que vous aurez enregistré le formulaire).",
            required=True,
        ),
    ],
)

contexte_contractuel = FieldSet(
    "contexte_contractuel",
    "Contexte contractuel et propriété intellectuelle",
    [
        BooleanField(
            "contrats",
            "Les travaux de recherche à l’origine de l’œuvre ont-ils été conduits dans le cadre de contrat ou partenariat de recherche ?",
            note="Par exemple financement Sorbonne Université et/ou CNRS, partenariat académique, contrat de recherche avec un partenaire privé, sponsoring, CIFRE, ANR, contrat européen, financement industriel, ministère, régional, OSEO, association, fondation, chaire…",
            required=True,
        ),
        ListeContrats(
            "liste_contrats", "Liste des contrats/partenariats", required=True
        ),
        #
        BooleanField(
            "autres_declarations",
            "L’œuvre est-elle liée à une autre (ou plusieurs) déclaration(s) d’invention ou de logiciel/base de données?",
            required=True,
        ),
        ListeAutresDeclarations(
            "liste_autres_declarations", "Liste des autres déclarations", required=True
        ),
        #
        BooleanField(
            "licences_existantes",
            "Y-a-t-il des licences existantes (licences propriétaires, licences libres, nom de la licence et n° de version de la licence)",
            required=True,
        ),
        ListeLicencesExistantes(
            "liste_licences_existantes", "Liste des licences existantes", required=True
        ),
        # TODO
    ],
)

pieces_a_joindre = FieldSet(
    "pieces_a_joindre",
    "Pièces à joindre",
    [
        HTML(
            """
    <p>Ce formulaire doit être accompagné de la "fiche individuelle de contributeur", remplie par chaque contributeur.</p>

    <p><a href="http://www.upmc.fr/modules/resources/download/default/espace_personnels/DGRTT/TRANSFERER/FICHE_INDIVIDUELLE_CONTRIBUTEUR_LOGICIEL_160915.docx">Téléchargez la fiche</a>, remplissez-la et envoyez-la à chaque contributeur afin qu'il la remplisse aussi et la signe.</p>

    <p>Une fois que vous avez enregistré le formulaire, utilisez l'onglet <b>Pièces à joindre</b> pour accompagner votre demande des documents suivants :</p>

    <ul>
        <li>une fiche individuelle de contributeur par contributeur</li>
    </ul>

    <p>et si nécessaire :</p>

    <ul>
        <li>un schéma de l'architecture du logiciel</li>
        <li>une copie des publications</li>
        <li>une copie du projet de communication</li>
    </ul>
    """
        )
    ],
)

engagement = FieldSet(
    "engagement",
    "Engagement de l'interlocuteur privilégié de l'invention",
    [
        Boolean2Field(
            "engagement",
            "En tant que porteur du projet, je reconnais accepter d’être le contact scientifique pour l’œuvre auprès de Sorbonne Université, et de la SATT LUTECH le cas échéant, au nom de l’ensemble des contributeurs, et confirme la validité des renseignements indiqués dans cette déclaration de logiciel/base de données.",
            required=True,
        )
    ],
)


#
# The Form
#
class DemandePiLogicielForm(Form):
    name = "pi_logiciel"
    template = "newforms/newform.html"
    label = "Déclaration de logiciel / base de données"

    fieldsets = [
        laboratoire,
        porteur_de_projet,
        description_oeuvre,
        type_oeuvre,
        description_technique,
        divulgation,
        contexte_contractuel,
        exploitation,
        engagement,
        pieces_a_joindre,
    ]

    def post_init(self):
        """Populate form messages from json constants."""
        conditions = get_constant("pi_logiciel.conditions")
        self.conditions = conditions



from __future__ import annotations

from collections import OrderedDict
from typing import Dict, Text


class Group:
    def __init__(self, id, label, fields):
        self.id = id
        self.label = label
        self.fields = fields


class Groups:
    def __init__(self):
        self.dict = OrderedDict()  # type: Dict[Text, Group]

    def add_group(self, id, label, fields):
        self.dict[id] = Group(id, label, fields)

    def __getitem__(self, item):
        return self.dict[item]

    def __iter__(self):
        return iter(self.dict)


class Struct:
    """See: http://stackoverflow.com/questions/16327141/"""

    def __init__(self, **entries):
        self.__dict__.update(entries)



from __future__ import annotations

import functools

from diskcache.core import ENOVAL, args_to_key, full_name
from flask import g


def memoize():
    def decorator(func):
        """Decorator created by memoize() for callable `func`."""
        base = (full_name(func),)

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            """Wrapper for callable to cache arguments and return values."""
            key = wrapper.__cache_key__(*args, **kwargs)
            result = g.cache.get(key, ENOVAL)

            if result is ENOVAL:
                result = func(*args, **kwargs)
                g.cache[key] = result

            return result

        def __cache_key__(*args, **kwargs):
            """Make key for cache given function arguments."""
            return args_to_key(base, args, kwargs, False)

        wrapper.__cache_key__ = __cache_key__
        return wrapper

    return decorator



# from __future__ import annotations
#
# from typing import Set
# from uuid import uuid4
#
# from labster.domain2.model.structure import Structure, StructureRepository
#
# from .base import BaseInMemoryRepository
#
#
# class InmemoryStructureRepository(BaseInMemoryRepository, StructureRepository):
#     def put(self, structure: Structure):
#         if not structure.id:
#             structure.id = str(uuid4())
#         self._data[structure.id] = structure
#
#     def delete(self, structure: Structure):
#         del self._data[structure.id]
#
#     def get_by_id(self, id: str) -> Structure:
#         return self._data[id]
#
#     def get_all(self) -> Set[Structure]:
#         return set(self._data.values())
from __future__ import annotations



from __future__ import annotations

from abilian.core.entities import Entity
from sqlalchemy import Column, Unicode


class Financeur(Entity):
    __indexable__ = False

    nom = Column(Unicode, nullable=False)
    sigle = Column(Unicode, nullable=False)
    type = Column(Unicode, nullable=False)
    sous_type = Column(Unicode, nullable=False)
    classe = Column(Unicode)
    pays = Column(Unicode, nullable=False)



"""Modèle de case management adapté au projet Labster."""
from __future__ import annotations

from datetime import date, datetime
from typing import TYPE_CHECKING, Container

import structlog

from labster.domain2.services.notifications import send_email
from labster.domain2.services.roles import Role
from labster.lib.workflow import Transition, Workflow

from .forms import get_form
from .states import ABANDONNEE, ACTIVE_STATES, EN_EDITION, EN_INSTRUCTION, \
    EN_VALIDATION, EN_VERIFICATION, REJETEE, TRAITEE

if TYPE_CHECKING:
    from labster.domain2.model.demande import Demande
    from labster.domain2.model.profile import Profile

logger = structlog.get_logger()


#
# Transitions
#
class BaseTransition(Transition):
    def get_form(self, workflow, **kw):
        return get_form()


class Abandonner(BaseTransition):
    label = "Abandonner la demande"
    category = "danger"
    from_states = [EN_EDITION, EN_VALIDATION, EN_VERIFICATION, EN_INSTRUCTION]
    to_state = ABANDONNEE
    message = "{actor} a abandonné la demande."

    def precondition(self, workflow: Workflow):
        return workflow.actor_is_porteur_or_gestionnaire()

    def get_users_to_notify(self, workflow, old_state):
        case = workflow.case

        if old_state == EN_EDITION:
            return set()

        if old_state == EN_VALIDATION:
            # FIXME
            # return case.owners + case.structure.direction
            return case.owners

        if old_state in (EN_VERIFICATION, EN_INSTRUCTION):
            return case.owners | {case.contact_labco}

        raise RuntimeError(f"Unknown state: {old_state}")


class Desarchiver(BaseTransition):
    label = "Désarchiver la demande"
    category = "danger"
    from_states = [ABANDONNEE, REJETEE, TRAITEE]
    message = "{actor} a désarchivé la demande."

    def apply(self, workflow, data):
        demande: Demande = workflow.case
        old_state_id = demande.wf_history[-1].get("old_state", "EN_EDITION")
        old_state = workflow.get_state_by_id(old_state_id)
        old_state.enter(workflow)
        demande.active = True
        demande.editable = True

    def get_form(self, workflow, **kw):
        return get_form(require_note=True)

    def get_users_to_notify(self, workflow, old_state):
        case = workflow.case
        return case.owners


class Soumettre(BaseTransition):
    label = "Soumettre la demande"
    from_states = [EN_EDITION]

    def precondition(self, workflow):
        demande = workflow.case
        structure_de_la_demande = demande.structure
        actor = workflow.actor

        if not demande.is_valid():
            return False

        # Le porteur peut toujours soumettre
        if actor == demande.porteur:
            return True

        if (
            actor.has_role(Role.GESTIONNAIRE, structure_de_la_demande)
            and structure_de_la_demande.permettre_soummission_directe
        ):
            return True

        if actor == demande.gestionnaire and workflow.get_value("validee_hierarchie"):
            return True

        return False

        #
        # TODO: vérifier si c'est OK.
        #
        # Cas à gérer: la demande est issue d'une sous-structure dans
        # le périmètre du gestionnaire.
        #
        # demande.structure et workflow.actor.structure peuvent être différentes.
        #
        # Par exemple:
        # demande.structure = equipe Developpement des circuits neuronaux
        # workflow.actor.structure = Institut de la vision
        # Et dans ce cas
        # demande.structure.get_gestionnaires() = []
        #
        # Sur le formulaire de la demande, l'utilisateur voit
        # "Institut de la vision", donc peut aller changer son paramètre
        # "permettre_soummission_directe" à True
        # On a donc
        # demande.structure.permettre_soummission_directe inchangé, potentiellement à False,
        # et workflow.actor.structure.permettre_soummission_directe à True.
        #
        # On a donc besoin des deux "if" suivants.

        # if structure_de_l_acteur.permettre_soummission_directe:
        #     if (
        #         actor in structure_de_l_acteur.get_gestionnaires()
        #     ):
        #         return True
        #
        #     if (
        #         actor in structure_de_la_demande.get_gestionnaires()
        #     ):
        #         return True
        #
        #     if (
        #         workflow.get_value("validee_hierarchie")
        #         and actor == demande.gestionnaire
        #     ):
        #         return True
        #
        # return False

    def apply(self, workflow, data):
        demande = workflow.case
        demande.editable = False

        if data.get("resoumission"):
            workflow.set_value("validee_hierarchie", False)
            workflow.set_value("signatures", [])

        if workflow.get_value("validee_hierarchie"):
            if workflow.get_value("recevable"):
                EN_INSTRUCTION.enter(workflow)
            else:
                EN_VERIFICATION.enter(workflow)
        else:
            EN_VALIDATION.enter(workflow)

    def message(self, workflow):
        if workflow.get_value("validee_hierarchie"):
            return "{actor} a resoumis sa demande sans revalidation hiérarchique"
        else:
            return "{actor} a soumis sa demande pour validation hiérarchique."

    def get_users_to_notify(self, workflow, old_state):
        case = workflow.case

        if workflow.get_value("validee_hierarchie"):
            return {case.contact_labco}

        structure = case.structure
        result = structure.responsables
        for structure_concernee in case.get_structure_concernees():
            result |= structure_concernee.responsables

        return result

    def get_form(self, workflow, **kw):
        ask_for_revalidation = workflow.get_value("validee_hierarchie")
        return get_form(ask_for_revalidation=ask_for_revalidation)


#
# Prendre la main
#
class AbstractPrendreLaMain(BaseTransition):
    label = "Prendre la main sur la demande"
    category = "danger"
    message = "{actor} a pris la main sur la demande."

    def get_users_to_notify(self, workflow, old_state):
        case = workflow.case
        actor = workflow.actor
        user_ids = {entry["actor_id"] for entry in case.wf_history}
        users = set()
        for id in user_ids:
            try:
                # TODO
                user = Profile.query.get(id)
                if user != actor:
                    users.add(user)
            except Exception:
                pass
        return users


class PrendreLaMainGestionnaire(AbstractPrendreLaMain):
    from_states = ACTIVE_STATES

    def precondition(self, workflow: Workflow):
        from labster.rbac import is_gestionnaire

        case = workflow.case
        actor = workflow.actor

        return is_gestionnaire(actor, case) and actor not in (
            case.gestionnaire,
            case.porteur,
        )

    def apply(self, workflow, data):
        case = workflow.case
        actor = workflow.actor
        case.gestionnaire = actor


class PrendreLaMainDgrtt(AbstractPrendreLaMain):
    from_states = ACTIVE_STATES

    def precondition(self, workflow):
        from labster.rbac import is_membre_drv

        case = workflow.case
        actor = workflow.actor

        if actor == case.contact_labco:
            return False

        if actor.is_membre_dri():
            return True

        if is_membre_drv(actor, case.structure):
            return True

        return False

    def apply(self, workflow, data):
        case = workflow.case
        actor = workflow.actor
        case.contact_labco = actor


#
# "Valider demande
#
class ValiderDir(BaseTransition):
    label = "Valider la demande"
    from_states = [EN_VALIDATION]
    message = "Demande validée par un responsable ({actor})."

    def precondition(self, workflow):
        return workflow.actor in workflow.state.task_owners(workflow)

    def apply(self, workflow, data):
        demande = workflow.case
        actor = workflow.actor

        signatures: set[str] = set(workflow.get_value("signatures") or [])

        structures_signataires = demande.structures_signataires()
        for structure in structures_signataires:
            if actor.has_role(Role.RESPONSABLE, structure):
                signatures.add(structure.id)

        workflow.set_value("signatures", list(signatures))

        structures_signataires_restantes = {
            s for s in structures_signataires if s.id not in signatures
        }

        if not structures_signataires_restantes:
            demande.assigne_contact_labco()
            workflow.set_value("validee_hierarchie", True)
            demande.date_effective = date.today()
            EN_VERIFICATION.enter(workflow)
            return

    def get_users_to_notify(self, workflow, old_state) -> set[Profile]:
        case = workflow.case
        if case.contact_labco:
            return case.owners | {case.contact_labco}
        else:
            return case.owners


class RequerirModificationDir(BaseTransition):
    label = "Requérir modification / complément"
    from_states = [EN_VALIDATION]
    to_state = EN_EDITION
    message = (
        "Demande de compléments / modifications par {actor} "
        "(responsable labo/département/équipe) pour vérification de recevabilité."
    )

    def precondition(self, workflow):
        return workflow.actor in workflow.state.task_owners(workflow)

    def apply(self, workflow, data):
        workflow.set_value("validee_hierarchie", False)
        workflow.set_value("signatures", [])

    def get_users_to_notify(self, workflow, old_state):
        case = workflow.case
        return case.owners

    def get_form(self, workflow, **kw):
        return get_form(require_note=True)


#
# DGRTT (renommé DR&I)
#
class AccuserReception(BaseTransition):
    label = "Accuser réception en attendant vérification ultérieure"
    from_states = [EN_VERIFICATION]
    to_state = EN_VERIFICATION
    message = "Accusé de réception envoyé par {actor} (contact)."

    def precondition(self, workflow):
        return workflow.actor_is_contact_labco() and not workflow.get_value(
            "ar_envoye", False
        )

    def apply(self, workflow, data):
        workflow.set_value("ar_envoye", True)

    def get_users_to_notify(self, workflow, old_state):
        case = workflow.case
        return case.owners


class RequerirModificationDgrtt(BaseTransition):
    label = "Requérir modification / complément"
    from_states = [EN_VERIFICATION, EN_INSTRUCTION]
    to_state = EN_EDITION

    def precondition(self, workflow):
        return workflow.actor_is_contact_labco()

    def apply(self, workflow, data):
        if data.get("resoumission"):
            workflow.set_value("validee_hierarchie", False)
            workflow.set_value("signatures", [])

    def message(self, workflow):
        if workflow.state == EN_VERIFICATION:
            return (
                "Demande de compléments / modifications par {actor} "
                "(contact) pour vérification de recevabilité."
            )
        else:
            return (
                "Demande de compléments / modifications par {actor} "
                "(contact) pour instruction."
            )

    def get_users_to_notify(self, workflow, old_state):
        case = workflow.case
        return case.owners

    def get_form(self, workflow, **kw):
        ask_for_revalidation = workflow.get_value("validee_hierarchie")
        return get_form(ask_for_revalidation=ask_for_revalidation, require_note=True)


class ConfirmerRecevabiliteDgrtt(BaseTransition):
    label = "Confirmer recevabilité"
    from_states = [EN_VERIFICATION]
    to_state = EN_INSTRUCTION

    def precondition(self, workflow):
        return workflow.actor_is_contact_labco()

    def apply(self, workflow, data):
        workflow.set_value("recevable", True)
        self.send_notification(workflow)

    def send_notification(self, workflow):
        case = workflow.case
        subject = "Recevabilité de votre demande par la DR&I"
        ctx = {
            "transition": self,
            "demande": case,
            "workflow": workflow,
            "now": datetime.now(),
        }
        send_email(case.owners, subject, "notif-demande-recevable.html", ctx)

    def get_users_to_notify(self, workflow, old_state):
        case = workflow.case
        return case.owners

    def get_form(self, workflow, **kw):
        return get_form(with_no_infolab=True)

    def message(self, workflow):
        demande = workflow.case
        tpl = (
            "Recevabilité confirmée par {actor} (contact). "
            "No Infolab: %s." % demande.no_infolab
        )
        return tpl


class ConfirmerFinalisationDgrtt(BaseTransition):
    label = "Confirmer finalisation"
    from_states = [EN_INSTRUCTION]
    to_state = TRAITEE

    def precondition(self, workflow):
        return workflow.actor_is_contact_labco()

    def get_users_to_notify(self, workflow, old_state):
        case = workflow.case
        return case.owners

    def get_form(self, workflow, **kw):
        return get_form(with_no_eotp=True)

    def apply(self, workflow, data):
        self.send_notification(workflow)
        # self.generate_docs(workflow.case)

    def send_notification(self, workflow):
        case = workflow.case
        ctx = {
            "transition": self,
            "demande": case,
            "workflow": workflow,
            "now": datetime.now(),
        }
        subject = "Finalisation de votre demande par la DR&I"
        du = case.structure.responsables
        recipients = case.owners | du
        send_email(recipients, subject, "notif-demande-finalisee.html", ctx)

    def message(self, workflow):
        demande = workflow.case
        tpl = (
            "Traitement finalisé par {actor} (contact)."
            "No eOTP: %s." % demande.no_eotp
        )
        return tpl

    # Warning: not used anymore currently
    # def generate_docs(self, demande: Demande):
    #     from labster.di import injector
    #     from labster.domain2.model.demande import DemandeRH
    #
    #     db = injector.get(SQLAlchemy)
    #
    #     if not isinstance(demande, DemandeRH):
    #         return
    #
    #     docs = list(demande.documents_generes or [])
    #
    #     def add_pj(b: bytes, name: str):
    #         from flask import current_app
    #
    #         if current_app.config.get("TESTING"):
    #             return
    #
    #         blob = Blob(b)
    #         db.session.add(blob)
    #         db.session.flush()
    #
    #         d = {
    #             "name": name,
    #             "date": datetime.utcnow().isoformat(),
    #             "blob_id": blob.id,
    #         }
    #         docs.append(d)
    #
    #     add_pj(devis_rh(demande), "Devis RH.pdf")
    #     add_pj(lettre_commande_rh(demande), "Lettre de Commande.pdf")
    #
    #     demande.documents_generes = docs


class RejeterDgrtt(BaseTransition):
    label = "Rejeter / abandonner demande"
    category = "danger"
    from_states = [EN_EDITION, EN_VALIDATION, EN_VERIFICATION, EN_INSTRUCTION]
    to_state = REJETEE
    message = "Demande rejetée / abandonnées par {actor} (contact)."

    def precondition(self, workflow):
        return workflow.actor_is_contact_labco()

    # def precondition(self, workflow):
    #     from labster.rbac import is_membre_dri, is_membre_drv
    #
    #     actor: Profile = workflow.actor
    #     case: Demande = workflow.case
    #     if is_membre_dri(actor):
    #         return True
    #
    #     if is_membre_drv(actor, case.structure):
    #         return True
    #
    #     for structure in case.get_structure_concernees():
    #         if is_membre_drv(actor, structure):
    #             return True
    #
    #     return False

    def get_users_to_notify(self, workflow, old_state):
        case = workflow.case
        return case.owners

    def apply(self, workflow, data):
        self.send_notification(workflow)

    def send_notification(self, workflow: Workflow):
        case = workflow.case
        ctx = {
            "transition": self,
            "demande": case,
            "workflow": workflow,
            "now": datetime.now(),
        }
        subject = "Rejet de votre demande par la DR&I ou la DRV"
        send_email(case.owners, subject, "notif-demande-rejetee.html", ctx)

    def get_form(self, workflow, **kw):
        return get_form(require_note=True)


class Commenter(BaseTransition):
    label = "Envoyer un message"
    category = "success"
    from_states = [EN_EDITION, EN_VERIFICATION, EN_INSTRUCTION]
    message = "{actor} a posté le commentaire ou la question suivante: "

    def precondition(self, workflow: Workflow):
        actor = workflow.actor
        demande = workflow.case
        if workflow.state == EN_EDITION and not demande.contact_labco:
            return False
        stakeholders = self._get_stakeholders(workflow)
        return actor in stakeholders

    def get_users_to_notify(self, workflow, old_state):
        return self._get_stakeholders(workflow)

    def _get_stakeholders(self, workflow: Workflow) -> Container:
        from flask_sqlalchemy import SQLAlchemy

        from labster.di import injector
        from labster.domain2.model.profile import Profile

        db = injector.get(SQLAlchemy)

        demande: Demande = workflow.case
        stakeholders = set(demande.owners)
        if demande.contact_labco:
            stakeholders.add(demande.contact_labco)

        # FIXME
        # structure: Structure = demande.structure
        # directeurs = structure.get_directeurs()
        # for directeur in directeurs:
        #     stakeholders.add(directeur)

        for history_item in demande.wf_history:
            actor_id = history_item["actor_id"]
            if actor_id:
                # TODO
                if isinstance(actor_id, int):
                    actor = db.session.query(Profile).filter_by(old_id=actor_id).first()
                else:
                    actor = db.session.query(Profile).get(actor_id)
                if actor:
                    stakeholders.add(actor)

        return stakeholders

    def get_form(self, workflow, **kw):
        return get_form(require_note=True)

    def apply(self, workflow: Workflow, data: dict):
        self.send_notification(workflow)

    def send_notification(self, workflow: Workflow) -> None:
        case = workflow.case
        ctx = {
            "transition": self,
            "demande": case,
            "workflow": workflow,
            "now": datetime.now(),
        }
        subject = "Un commentaire sur votre demande"
        send_email(case.owners, subject, "notif-demande-comment.html", ctx)


#
ABANDONNER = Abandonner()
DESARCHIVER = Desarchiver()
SOUMETTRE = Soumettre()
PRENDRE_LA_MAIN_GESTIONNAIRE = PrendreLaMainGestionnaire()
PRENDRE_LA_MAIN_DGRTT = PrendreLaMainDgrtt()
#
VALIDER_DIR = ValiderDir()
REQUERIR_MODIFICATION_DIR = RequerirModificationDir()
#
ACCUSER_RECEPTION = AccuserReception()
REQUERIR_MODIFICATION_DGRTT = RequerirModificationDgrtt()
CONFIRMER_RECEVABILITE_DGRTT = ConfirmerRecevabiliteDgrtt()
CONFIRMER_FINALISATION_DGRTT = ConfirmerFinalisationDgrtt()
REJETER_DGRTT = RejeterDgrtt()
#
COMMENTER = Commenter()



from __future__ import annotations

from flask_mail import Mail
from flask_sqlalchemy import SQLAlchemy
from injector import Injector, singleton
from sqlalchemy.orm import Session

from labster.auth import AuthContext
from labster.domain2.model.demande import DemandeRepository
from labster.domain2.model.profile import ProfileRepository
from labster.domain2.model.structure import StructureRepository
from labster.domain2.services.contacts import ContactService
from labster.domain2.services.roles import RoleService
from labster.extensions import db, mail
from labster.infrastructure.repositories.sqla import SqlaProfileRepository
from labster.infrastructure.repositories.sqla.demande_repository import \
    SqlaDemandeRepository
from labster.infrastructure.repositories.sqla.structure_repository import \
    SqlaStructureRepository
from labster.infrastructure.services.sqla.contacts import SqlaContactService
from labster.infrastructure.services.sqla.roles import SqlaRoleService


def configure(binder):
    binder.bind(ProfileRepository, to=SqlaProfileRepository, scope=singleton)
    binder.bind(StructureRepository, to=SqlaStructureRepository, scope=singleton)
    binder.bind(RoleService, to=SqlaRoleService, scope=singleton)
    binder.bind(ContactService, to=SqlaContactService, scope=singleton)
    binder.bind(DemandeRepository, to=SqlaDemandeRepository, scope=singleton)

    binder.bind(SQLAlchemy, to=db, scope=singleton)
    # binder.bind(Mapper, scope=singleton)
    binder.bind(Session, to=db.session, scope=singleton)

    binder.bind(Mail, to=mail, scope=singleton)
    binder.bind(AuthContext, scope=singleton)


injector = Injector(modules=[configure])



from __future__ import annotations

from pathlib import Path
from typing import Callable

from werkzeug.utils import import_string

registry: dict[str, Callable] = {}


def context_for(name: str) -> Callable:
    def wrap(f):
        f._exposed = True
        registry[name] = f
        return f

    return wrap


def register_submodules():
    here = "labster.rpc"
    subdirs = ["commands", "queries", "bi"]

    import_string(f"{here}.context")

    this_dir = Path(__file__).parent
    for subdir in subdirs:
        for path in (this_dir / subdir).glob("*.py"):
            name = path.parts[-1][0:-3]
            if name == "__init__":
                continue

            fqn = f"{here}.{subdir}.{name}"
            import_string(fqn)



from __future__ import annotations

from labster.newforms.rh import DemandeRHForm


def test_export_empty_rh_form_as_dict(db_session):
    form = DemandeRHForm()
    exported = form.to_dict()
    assert "fieldsets" in exported
    fieldsets = exported["fieldsets"]
    assert any(lambda x: x["name"] == "nature" for x in fieldsets)


def test_export_rh_form_with_data_as_dict(db_session):
    model = {"type_de_demande": "Renouvellement"}
    form = DemandeRHForm(model=model)
    exported = form.to_dict()
    assert "fieldsets" in exported



"""De l'ancien annuaire:

260 {LL}Labo homologué (en cotutelle ou non)
 241 {SP}Service propre à l'UPMC
 238 {DO}Direction opérationnelle
 226 {SE}Service d'enseignement
  67 {C}Conseil ou commission d'université
  58 {SX}Service extérieur à l'UPMC
  58 {SS}Équipe, bureau, sous-service
  53 {LX}Laboratoire extérieur à l'UPMC
  21 {ED}Ecole Doctorale
  18 {IF}Institut Fédératif de Recherche
  16 {LF}Fédération de Recherche
  13 {UD}Unités extérieures diverses
   8 {LE}Équipe de recherche à l'UPMC
   8 {IO}Institut ou Observatoire
   8 {CU}Unité de Formation et de Recherche
   6 {SC}Service Commun
   3 {DI}Directoire d'université
   2 {CX}Ensemble de services ext UPMC
   1 {D}Decret
"""
from __future__ import annotations

import math
from abc import ABC, ABCMeta
from pprint import pformat
from typing import TYPE_CHECKING, Any, Collection
from uuid import uuid4

from abilian.app import db
from sqlalchemy import Boolean, Column, ForeignKey, Integer, String, Table
from sqlalchemy.orm import backref, relationship

from .base import Repository
from .type_structure import ED, TypeStructure, get_type_structure

if TYPE_CHECKING:
    from labster.domain2.services.roles import Role

    from .profile import Profile


hierarchy = Table(
    "v3_hierarchy",
    db.metadata,
    Column("parent_id", String(36), ForeignKey("v3_structures.id")),
    Column("child_id", String(36), ForeignKey("v3_structures.id")),
)


class StructureId(str):
    @staticmethod
    def new():
        return StructureId(uuid4())


# @attrs(eq=False, order=False, repr=False, auto_attribs=True)
class Structure(db.Model):
    # Cf. Slides de specs, page 2.
    """Un structure S comprend:

    • Un nom
    • Un acronyme (éventuellement vide)
    • Un type (cf. tableau page suivante)

    • Un champ LDAP, pour certaines structures réelles qui seront importées depuis l’annuaire
    • Une liste de structures parentes (ascendants directs)
    • Une liste de structures filles (descendantes directes)

    + Utilisateurs (générés par les rôles):

    • Une liste d’utilisateurs (dits « rattachés » si S est de type réel ou « participants » sinon)
    • Une R liste d’utilisateurs (rôle) dits « responsables de S
      » (typiquement R correspondant au rôle Directeur/Direction ;
      correspondant au rôle Président/Présidence dans le cas où
      S est Sorbonne Université et au rôle Doyen/Décanat dans le
      cas où S est l’une des trois facultés) qui sont :
        – Rattachés à S si S est réelle
        – Rattachés à l’un des laboratoires « enfants » si S est un regroupement de laboratoires
        – Rattachés au laboratoire dont ils font partie, si S est un département ou une équipe
    • Un utilisateur dit « signataire »
    • Une liste de gestionnaires
    • Une liste d’administrateurs
    """

    __tablename__ = "v3_structures"

    id = Column(String(36), primary_key=True)
    old_id = Column(Integer)
    active = Column(Boolean, default=True, nullable=False)
    #: Un type (cf. tableau page suivante)
    type_name = Column(String)

    #: Nom
    nom = Column(String, default="", nullable=False)
    #: Un acronyme (éventuellement vide)
    sigle = Column(String, default="", nullable=False)
    #: Un champ LDAP, pour les structures réelles
    dn = Column(String, default="", nullable=False)
    #: Old DN, in case this comes from the old directory
    old_dn = Column(String, default="", nullable=False)
    #: Nouvel id LDAP
    supann_code_entite = Column(String)

    #: Propriétés additionnelles (non-spécifiées)
    email = Column(String, default="", nullable=False)
    permettre_reponse_directe = Column(Boolean, default=True)
    permettre_soummission_directe = Column(Boolean, default=False)

    children = relationship(
        "Structure",
        secondary=hierarchy,
        primaryjoin=(lambda: hierarchy.c.parent_id == Structure.id),
        secondaryjoin=(lambda: hierarchy.c.child_id == Structure.id),
        collection_class=set,
        backref=backref("parents", collection_class=set),
    )

    def __init__(self, **kw):
        self.id = str(uuid4())

        self.active = True
        self._depth = -1
        self.permettre_reponse_directe = True
        self.permettre_soummission_directe = False
        self.email = ""
        self.sigle = ""
        self.dn = ""
        self.old_dn = ""

        super().__init__(**kw)

    def __str__(self):
        return f"<Structure type='{self.type_name}' nom='{self.nom}' id={id(self)}>"

    def __repr__(self):
        return str(self)

    @property
    def name(self) -> str:
        return self.nom

    def get_type(self) -> TypeStructure:
        return get_type_structure(self.type_name)

    def set_type(self, type: TypeStructure) -> None:
        self.type_name = type.name

    type = property(get_type, set_type)

    @property
    def is_reelle(self) -> bool:
        return self.type.reel

    @property
    def sigle_ou_nom(self) -> str:
        return self.sigle or self.nom

    @property
    def parent(self) -> Structure | None:
        if self.real_parents:
            return list(self.real_parents)[0]
        elif self.parents:
            return list(self.parents)[0]
        else:
            return None

    @property
    def real_parents(self) -> Collection[Structure]:
        return {p for p in self.parents if p.is_reelle}

    @property
    def virtual_parents(self) -> Collection[Structure]:
        return {p for p in self.parents if not p.is_reelle}

    def add_child(self, child: Structure):
        if not self.can_have_child(child):
            raise ValueError(f"Forbidden relation: {self}->{child}")

        child._depth = -1
        # child.parents.add(self)
        self.children.add(child)

    def remove_child(self, child: Structure):
        child._depth = -1
        child.parents.remove(self)
        # self.children.remove(child)

    def can_have_child(self, other: Structure) -> bool:
        if other.parents and not other.type.can_have_multiple_parents:
            return False

        my_type = self.type
        other_type = other.type

        return my_type.can_have_child_of_type(other_type)

    def add_parent(self, parent: Structure):
        parent.add_child(self)

    def remove_parent(self, parent: Structure):
        parent.remove_child(self)

    def can_have_parent(self, other: Structure) -> bool:
        return other.can_have_child(self)

    def delete(self):
        assert not self.children
        for parent in list(self.parents):
            self.remove_parent(parent)
        self.active = False

    @property
    def ancestors(self) -> list[Structure]:
        parent = self.parent
        if not parent:
            return []
        else:
            return [parent] + parent.ancestors

    @property
    def descendants(self) -> set[Structure]:
        result = set()
        for child in self.children:
            result.add(child)
            result.update(child.descendants)
        return result

    @property
    def depth(self) -> int:
        if getattr(self, "_depth", -1) > -1:
            return self._depth
        parent = self.parent
        result = 0
        while True:
            if not parent:
                self._depth = result
                return result
            result += 1
            parent = parent.parent

    @property
    def path(self):
        result = []
        for ancestor in reversed(self.ancestors):
            result.append(ancestor.name)
        return result

    #
    # Roles
    #
    @property
    def responsables(self) -> set[Profile]:
        from labster.domain2.services.roles import Role

        return self._get_users_with_role(Role.RESPONSABLE)

    @property
    def gestionnaires(self) -> set[Profile]:
        from labster.domain2.services.roles import Role

        return self._get_users_with_role(Role.GESTIONNAIRE)

    @property
    def signataire(self) -> Profile | None:
        from labster.domain2.services.roles import Role

        signataires = self._get_users_with_role(Role.SIGNATAIRE)
        if signataires:
            return signataires.pop()
        else:
            return None

    def _get_users_with_role(self, role: Role) -> set[Profile]:
        from labster.di import injector
        from labster.domain2.services.roles import RoleService

        role_service = injector.get(RoleService)
        responsables = role_service.get_users_with_given_role(role, self)
        return responsables

    #
    # Misc
    #
    def check(self):
        state = vars(self)
        try:
            for k, v in state.items():
                if isinstance(v, float) and math.isnan(v):
                    raise ValueError(f"Attribute {k} is NaN")

            self.type.full_check(self)

            if self.type == ED:
                assert len(self.parents) == 1
                assert self.parent.sigle == "IFD"

        except (AssertionError, ValueError):
            msg = f"Check failed on {self.sigle_ou_nom}. " + pformat(state)
            print(msg)
            raise

    validate = check


class StructureRepository(Repository, ABC, metaclass=ABCMeta):
    # @abstractmethod
    def put(self, structure: Structure):
        raise NotImplementedError

    # @abstractmethod
    def delete(self, structure: Structure):
        raise NotImplementedError

    # @abstractmethod
    def get_all(self) -> set[Structure]:
        raise NotImplementedError

    # Generic (slow) implem
    def get_by(self, key: str, value: Any) -> Structure | None:
        all_objects = self.get_all()
        for x in all_objects:
            if getattr(x, key) == value:
                return x
        return None

    def get_by_id(self, id: str) -> Structure | None:
        return self.get_by("id", id)

    def get_by_old_dn(self, old_dn: str) -> Structure | None:
        return self.get_by("old_dn", old_dn)

    def get_by_dn(self, dn: str) -> Structure | None:
        return self.get_by("dn", dn)

    def get_by_sigle(self, sigle: str) -> Structure | None:
        return self.get_by("sigle", sigle)

    def get_by_old_id(self, old_id: int) -> Structure | None:
        return self.get_by("old_id", old_id)

    def get_root(self) -> Structure:
        result = self.get_by_dn("ou=SU,ou=Affectations,dc=chapeau,dc=fr")
        assert result
        return result

    def check_all(self):
        for obj in self.get_all():
            obj.check()



from __future__ import annotations

import dateutil.parser
import pandas as pd
import ramda as r
from flask_sqlalchemy import SQLAlchemy
from jsonrpcserver import method
from werkzeug.exceptions import Forbidden

from labster.bi.model import StatsLine
from labster.di import injector
from labster.domain2.model.demande import DemandeType
from labster.domain2.model.profile import Profile
from labster.domain2.model.structure import Structure
from labster.domain2.services.roles import Role
from labster.domain2.services.workflow.states import ALL_STATES
from labster.rpc.registry import context_for
from labster.security import get_current_profile, get_current_user

from .form import get_selectors
from .util import mes_structures

SECONDS_IN_DAY = 60.0 * 60 * 24
ALLOWED_ROLES = ["alc", "directeur", "chef de bureau", "gouvernance", "direction dgrtt"]

db = injector.get(SQLAlchemy)


@method
@context_for("bi")
def get_bi_context():
    check_permission()

    user = get_current_profile()
    ctx = get_stats2(user)
    ctx["selectors"] = get_selectors()
    return ctx


@method
def get_stats(**args):
    check_permission()

    args2 = {}
    for arg_name, arg_value in args.items():
        if not arg_value:
            continue
        if isinstance(arg_value, list):
            value = r.map(lambda x: x["value"], arg_value)
        elif isinstance(arg_value, dict):
            value = arg_value["value"]
        else:
            value = arg_value
        args2[arg_name] = value

    user = get_current_profile()
    return get_stats2(user, **args2)


def check_permission():
    user = get_current_user()
    if user.is_anonymous:
        raise Forbidden()

    profile = user.profile
    allowed = False
    if profile.has_role(Role.RESPONSABLE, "*"):
        allowed = True
    if profile.has_role(Role.ADMIN_CENTRAL):
        allowed = True

    if not allowed:
        raise Forbidden()


def get_stats2(
    user: Profile,
    periode_debut=None,
    periode_fin=None,
    type_demande=None,
    type_recrutement=None,
    financeur=None,
    structure_id=None,
    porteur_id=None,
):

    query = StatsLine.query

    if periode_debut:
        query = query.filter(StatsLine.date_soumission >= periode_debut)

    if periode_fin:
        query = query.filter(StatsLine.date_soumission <= periode_fin)

    if type_demande:
        query = query.filter(StatsLine.type_demande.in_(type_demande))

    if type_recrutement:
        query = query.filter(StatsLine.type_recrutement.in_(type_recrutement))

    if financeur:
        query = query.filter(StatsLine.financeur.in_(financeur))

    if porteur_id not in ("", "None", None):
        query = query.filter(StatsLine.porteur_id == porteur_id)

    if structure_id not in ("", "None", None):
        if not user.has_role(Role.ADMIN_CENTRAL):
            structures = mes_structures(user)
            if structure_id not in {s.id for s in structures}:
                raise Forbidden()

        structure = db.session.query(Structure).get(structure_id)
        descendant_ids = [s.id for s in structure.descendants]
        query = query.filter(
            StatsLine.structure_id.in_([structure_id] + descendant_ids)
        )

    elif user.has_role(Role.RESPONSABLE, "*"):
        structures = mes_structures(user)
        query = query.filter(StatsLine.structure_id.in_([s.id for s in structures]))

    # Stats
    totals = {}
    for state in ALL_STATES:
        state_id = state.id.lower()
        totals["nb_" + state_id] = query.filter(StatsLine.wf_state == state.id).count()

    totals["nb_en_cours"] = sum(
        totals[id]
        for id in [
            "nb_en_edition",
            "nb_en_validation",
            "nb_en_verification",
            "nb_en_instruction",
        ]
    )
    totals["nb_archivee"] = sum(
        totals[id] for id in ["nb_traitee", "nb_rejetee", "nb_abandonnee"]
    )
    totals["nb_total"] = totals["nb_en_cours"] + totals["nb_archivee"]
    assert totals["nb_total"] == query.count()

    totals = {k: int(v) for k, v in totals.items()}

    stats = {
        "conventions": make_conventions_stats(query),
        "rh": make_rh_stats(query),
        "avenants": make_avenants_stats(query),
        "pi_logiciel": make_pi_logiciel_stats(query),
        "pi_invention": make_pi_invention_stats(query),
        "duree_traitement": make_duree_traitement_stats(query),
    }

    ctx = {
        "totals": totals,
        "stats": stats,
    }

    return ctx


def parse_date(s):
    return dateutil.parser.parse(s)


def make_rh_stats(query):
    lines = query.filter(StatsLine.type_demande == DemandeType.RECRUTEMENT.value).all()
    result = make_stats(lines, ["duree", "salaire_brut_mensuel", "cout_total_mensuel"])
    result["count"] = len(lines)
    return result


def make_conventions_stats(query):
    lines = query.filter(StatsLine.type_demande == DemandeType.CONVENTION.value).all()
    result = make_stats(lines, ["montant", "recrutements_prev", "duree"])
    result["count"] = len(lines)
    return result


def make_avenants_stats(query):
    lines = query.filter(
        StatsLine.type_demande == DemandeType.AVENANT_CONVENTION.value
    ).all()
    result = make_stats(lines, [])
    result["count"] = len(lines)
    return result


def make_pi_logiciel_stats(query):
    lines = query.filter(StatsLine.type_demande == DemandeType.PI_LOGICIEL.value).all()
    result = make_stats(lines, [])
    result["count"] = len(lines)
    return result


def make_pi_invention_stats(query):
    lines = query.filter(StatsLine.type_demande == DemandeType.PI_INVENTION.value).all()
    result = make_stats(lines, [])
    result["count"] = len(lines)
    return result


def make_duree_traitement_stats(query):
    def duree_traitement(line):
        if line.date_soumission and line.date_finalisation:
            dt = line.date_finalisation - line.date_soumission
            days = dt.days + dt.seconds / SECONDS_IN_DAY
            return days
        else:
            return None

    lines = query.all()
    series = pd.Series([duree_traitement(line) for line in lines])
    return r.map(
        format_float,
        [
            series.mean(),
            series.median(),
            series.std(),
            series.min(),
            series.max(),
            series.sum(),
        ],
    )


def make_stats(lines, attrs):
    result = {}
    for attr in attrs:
        series = pd.Series([getattr(line, attr) for line in lines])
        result[attr] = r.map(
            format_float,
            [
                series.mean(),
                series.median(),
                series.std(),
                series.min(),
                series.max(),
                series.sum(),
            ],
        )
    return result


def format_float(x):
    return f"{float(x):,.2f}".replace(",", "\u00A0").replace(".", ",")



from __future__ import annotations

from sqlalchemy.orm import scoped_session

from labster.domain2.model.structure import Structure
from labster.domain2.model.type_structure import UN
from labster.rpc.queries.structures import sg_get_structure


def test_structures(structure_repo, db_session: scoped_session):
    universite = Structure(nom="SU", type_name=UN.name)
    structure_repo.put(universite)

    result = sg_get_structure(universite.id)
    expected = {
        "_depth": -1,
        "active": True,
        "ancestors": [],
        "can_be_deleted": True,
        "children": [],
        "dn": "",
        "email": "",
        "id": universite.id,
        "is_reelle": True,
        "nom": "SU",
        "old_dn": "",
        "old_id": None,
        "parents": [],
        "permettre_reponse_directe": True,
        "permettre_soummission_directe": False,
        "permissions": {},
        "sigle": "",
        "supann_code_entite": None,
        "type_name": "Université",
    }
    assert result == expected



from __future__ import annotations

from glom import glom
from sqlalchemy.orm import scoped_session

from labster.domain2.model.profile import Profile
from labster.domain2.model.structure import Structure
from labster.domain2.model.type_structure import UN
from labster.domain2.services.contacts import ContactType
from labster.rpc.commands.contacts import update_contacts
from labster.rpc.queries.contacts import get_contacts


def test_contacts(structure_repo, profile_repo, db_session: scoped_session):
    universite = Structure(nom="SU", type_name=UN.name)
    structure_repo.put(universite)

    result = get_contacts(universite.id)
    assert len(result) == 8

    user = Profile(uid="toto")
    profile_repo.put(user)

    contacts = {ct.name: "" for ct in ContactType}
    update_contacts(universite.id, contacts)
    result = get_contacts(universite.id)
    assert len(result) == 8
    assert glom(result, ["uid"]) == [None] * 8

    contacts["CONTACT_ENTREPRISES"] = user.uid
    update_contacts(universite.id, contacts)
    result = get_contacts(universite.id)
    assert len(result) == 8

    result1 = glom(result, ["uid"])
    assert result1[0] == "toto"
    assert result1[1:8] == [None] * 7

    contacts = {ct.name: user.uid for ct in ContactType}
    update_contacts(universite.id, contacts)
    result = get_contacts(universite.id)
    assert len(result) == 8
    assert glom(result, ["uid"]) == ["toto"] * 8



""""""
from __future__ import annotations

from os.path import dirname
from typing import Any

from flask import render_template
from flask_weasyprint import HTML

from labster.di import injector
from labster.domain2.model.demande import DemandeRH
from labster.domain2.model.structure import StructureRepository
from labster.domain2.services.calculs_couts import get_ctx_for_demande

structure_repo = injector.get(StructureRepository)


def devis_rh(demande: DemandeRH):
    ctx = get_ctx_for_demande(demande)
    ctx.update(additional_ctx_for_demande(demande))

    ctx["css_rules"] = open(dirname(__file__) + "/../../static/print/devis.css").read()
    html = render_template("documents-generes/devis-rh.html", **ctx)
    pdf = HTML(string=html).write_pdf()
    return pdf


def lettre_commande_rh(demande: DemandeRH):
    ctx = get_ctx_for_demande(demande)
    ctx.update(additional_ctx_for_demande(demande))

    if demande.data.get("nature_du_recrutement") == "Doctorant":
        template = "documents-generes/lettre-commande-recrutement-doctorant.html"
    else:
        template = "documents-generes/lettre-commande-recrutement.html"
    html = render_template(template, **ctx)
    pdf = HTML(string=html).write_pdf()
    return pdf


def additional_ctx_for_demande(demande: DemandeRH):
    ctx: dict[str, Any] = {}

    structure_inconnue = {
        "nom": "Structure inconnue",
        "signataire": None,
    }
    ctx["structure_d_accueil"] = demande.structure or structure_inconnue

    d = demande.data.get("structure_financeuse")
    if d:
        id = d["value"]
        ctx["structure_financeuse"] = structure_repo.get_by_id(id)
    else:
        ctx["structure_financeuse"] = demande.structure

    structure_financeuse = ctx["structure_financeuse"]

    if structure_financeuse:
        ctx["signataire"] = structure_financeuse.signataire
    else:
        ctx["structure_financeuse"] = structure_inconnue
        ctx["signataire"] = None

    return ctx



"""API tests."""
from __future__ import annotations

from flask.testing import FlaskClient
from sqlalchemy.orm.scoping import scoped_session

from labster.di import injector
from labster.domain2.model.structure import Structure, StructureRepository
from labster.domain2.model.type_structure import FA, UN
from labster.test.test_web import login_as_dgrtt
from labster.util import url_for

structure_repo = injector.get(StructureRepository)


def test_menu(client: FlaskClient, db_session: scoped_session) -> None:
    login_as_dgrtt(client, db_session)

    r = client.get(url_for("v3.menu"))
    assert r.status_code == 200


# def test_demandes(client, db_session):
#     login_as_dgrtt(client, db_session)
#
#     r = client.get(url_for("v3.demandes"))
#     assert r.status_code == 200


def test_structures(client: FlaskClient, db_session: scoped_session) -> None:
    login_as_dgrtt(client, db_session)

    universite = Structure(
        nom="Sorbonne Université",
        type_name=UN.name,
        sigle="SU",
        dn="ou=SU,ou=Affectations,dc=chapeau,dc=fr",
    )
    fac_sciences = Structure(nom="Faculté des Sciences", type_name=FA.name)
    structure_repo.put(universite)
    structure_repo.put(fac_sciences)

    r = client.get(url_for("v3.all_structures"))
    assert r.status_code == 200

    structure_repo.clear()


def test_users(client: FlaskClient, db_session: scoped_session) -> None:
    login_as_dgrtt(client, db_session)

    r = client.get(url_for("v3.all_users"))
    assert r.status_code == 200



# from __future__ import annotations
#
# import dateutil.parser
# import pandas as pd
# from flask import g, render_template, request
# from werkzeug.exceptions import Forbidden
#
# from labster.blueprints.bi.forms import FilterForm
# from labster.blueprints.bi.util import mes_structures
# from labster.domain.models.demandes import AVENANT_CONVENTION, CONVENTION, \
#     PI_INVENTION, PI_LOGICIEL, RECRUTEMENT
# from labster.domain.models.profiles import Profile
# from labster.domain.models.unites import OrgUnit
# from labster.domain.models.workflow import ALL_STATES
#
# from . import route
# from .model import StatsLine
#
# SECONDS_IN_DAY = 60.0 * 60 * 24
# ALLOWED_ROLES = ["alc", "directeur", "chef de bureau", "gouvernance", "direction dgrtt"]
#
# #
# @route("/")
# def home():
#     user = g.current_user  # type: Profile
#
#     if user.is_anonymous or not user.has_role(ALLOWED_ROLES):
#         raise Forbidden()
#
#     form = FilterForm(request.args)
#     form.validate()
#     ctx = {"title": "Statistiques", "form": form}
#
#     # Filtrage
#     periode_debut = form.periode_debut.data
#     periode_fin = form.periode_fin.data
#     types_demande = form.types_demande.data
#     types_recrutement = form.types_recrutement.data
#     financeurs = form.financeurs.data
#     structure_id = form.structure.data
#     porteur_id = form.porteur.data
#
#     query = StatsLine.query
#     if periode_debut:
#         query = query.filter(StatsLine.date_soumission >= periode_debut)
#     if periode_fin:
#         query = query.filter(StatsLine.date_soumission <= periode_fin)
#     if types_demande:
#         query = query.filter(StatsLine.type_demande.in_(types_demande))
#     if types_recrutement:
#         query = query.filter(StatsLine.type_recrutement.in_(types_recrutement))
#     if financeurs:
#         query = query.filter(StatsLine.financeur.in_(financeurs))
#     if porteur_id not in ("", "None", None):
#         query = query.filter(StatsLine.porteur_id == porteur_id)
#
#     if structure_id not in ("", "None", None):
#         structure_id = int(structure_id)
#         if user.has_role("recherche"):
#             structures = mes_structures(user)
#             if structure_id not in {s.id for s in structures}:
#                 raise Forbidden()
#
#         structure = OrgUnit.query.get(structure_id)
#         descendant_ids = [s.id for s in structure.descendants()]
#         query = query.filter(
#             StatsLine.structure_id.in_([structure_id] + descendant_ids)
#         )
#
#     elif user.has_role("directeur"):
#         structures = mes_structures(user)
#         query = query.filter(StatsLine.structure_id.in_([s.id for s in structures]))
#
#     # Stats
#     for state in ALL_STATES:
#         state_id = state.id.lower()
#         ctx["nb_" + state_id] = query.filter(StatsLine.wf_state == state.id).count()
#
#     ctx["nb_en_cours"] = sum(
#         ctx[id]
#         for id in [
#             "nb_en_edition",
#             "nb_en_validation",
#             "nb_en_verification",
#             "nb_en_instruction",
#         ]
#     )
#     ctx["nb_archivee"] = sum(
#         ctx[id] for id in ["nb_traitee", "nb_rejetee", "nb_abandonnee"]
#     )
#     ctx["nb_total"] = ctx["nb_en_cours"] + ctx["nb_archivee"]
#     assert ctx["nb_total"] == query.count()
#
#     stats = {
#         "conventions": make_conventions_stats(query),
#         "rh": make_rh_stats(query),
#         "avenants": make_avenants_stats(query),
#         "pi_logiciel": make_pi_logiciel_stats(query),
#         "pi_invention": make_pi_invention_stats(query),
#         "duree_traitement": make_duree_traitement_stats(query),
#     }
#
#     ctx["stats"] = stats
#
#     return render_template("bi/home.html", **ctx)
#
#
# def parse_date(s):
#     return dateutil.parser.parse(s)
#
#
# def make_rh_stats(query):
#     lines = query.filter(StatsLine.type_demande == RECRUTEMENT).all()
#     result = make_stats(lines, ["duree", "salaire_brut_mensuel", "cout_total_mensuel"])
#     result["count"] = len(lines)
#     return result
#
#
# def make_conventions_stats(query):
#     lines = query.filter(StatsLine.type_demande == CONVENTION).all()
#     result = make_stats(lines, ["montant", "recrutements_prev", "duree"])
#     result["count"] = len(lines)
#     return result
#
#
# def make_avenants_stats(query):
#     lines = query.filter(StatsLine.type_demande == AVENANT_CONVENTION).all()
#     result = make_stats(lines, [])
#     result["count"] = len(lines)
#     return result
#
#
# def make_pi_logiciel_stats(query):
#     lines = query.filter(StatsLine.type_demande == PI_LOGICIEL).all()
#     result = make_stats(lines, [])
#     result["count"] = len(lines)
#     return result
#
#
# def make_pi_invention_stats(query):
#     lines = query.filter(StatsLine.type_demande == PI_INVENTION).all()
#     result = make_stats(lines, [])
#     result["count"] = len(lines)
#     return result
#
#
# def make_duree_traitement_stats(query):
#     def duree_traitement(line):
#         if line.date_soumission and line.date_finalisation:
#             dt = line.date_finalisation - line.date_soumission
#             days = dt.days + dt.seconds / SECONDS_IN_DAY
#             return days
#         else:
#             return None
#
#     lines = query.all()
#     series = pd.Series([duree_traitement(line) for line in lines])
#     return [
#         series.mean(),
#         series.median(),
#         series.std(),
#         series.min(),
#         series.max(),
#         series.sum(),
#     ]
#
#
# def make_stats(lines, attrs):
#     result = {}
#     for attr in attrs:
#         series = pd.Series([getattr(line, attr) for line in lines])
#         result[attr] = [
#             series.mean(),
#             series.median(),
#             series.std(),
#             series.min(),
#             series.max(),
#             series.sum(),
#         ]
#     return result
from __future__ import annotations



from __future__ import annotations

from typing import Any

import ramda as r
from marshmallow import Schema, fields
from werkzeug.exceptions import Forbidden

from labster.domain2.services.constants import get_constant
from labster.domain2.services.roles import Role
from labster.domain.models.faq import FaqEntry
from labster.rpc.registry import context_for
from labster.security import get_current_profile


@context_for("faq")
def get_faq() -> dict[str, list[Any]]:
    entries = FaqEntry.query.order_by(FaqEntry.view_count.desc()).all()
    groups = sorted(r.group_by(lambda e: e.category, entries).items())

    result = []
    for k, v in groups:
        v = sorted(v, key=lambda x: -(x.view_count or 0))
        data = FaqEntrySchema().dump(v, many=True).data
        result.append([k, data])

    return {"categories": result, "entries": []}


class FaqEntrySchema(Schema):
    id = fields.String()
    title = fields.String()
    body = fields.String()
    category = fields.String()

    view_count = fields.Method("get_view_count")

    def get_view_count(self, obj: FaqEntry):
        return obj.view_count or 0


#
# Admin
#
@context_for("faq_admin")
def faq_admin():
    check_user_can_edit()
    entries = FaqEntry.query.order_by(FaqEntry.id).all()
    data = FaqEntrySchema().dump(entries, many=True).data

    categories = get_constant("faq_categories", [])
    return {"entries": data, "categories": categories}


def check_user_can_edit():
    profile = get_current_profile()
    if profile.has_role(Role.ADMIN_CENTRAL) or profile.has_role(Role.FAQ_EDITOR):
        return

    raise Forbidden



from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Set, Text

from flask import flash
from sqlalchemy.orm import joinedload

from labster.domain.services.constants import get_constant
from labster.extensions import db

if TYPE_CHECKING:
    from labster.domain.models.profiles import Profile
    from labster.domain.models.unites import OrgUnit


class BureauDgrtt:
    def __init__(self, id):
        self.id = id
        self._nom = None

    def __repr__(self):
        return f"<BureauDgrtt id={self.id} nom={self.nom}>"

    @property
    def nom(self):
        return get_constant("nom_bureaux_dgrtt." + self.id)

    @staticmethod
    def from_id(id):
        for bureau in BUREAUX_DGRTT:
            if bureau.id == id:
                return bureau
        raise IndexError()


BUREAUX_DGRTT = [
    BureauDgrtt("ETT"),
    BureauDgrtt("CFE"),
    BureauDgrtt("CP"),
    BureauDgrtt("CT"),
    BureauDgrtt("PIJ"),
    BureauDgrtt("PI2"),
    BureauDgrtt("REF"),
]


def bureaux():
    return BUREAUX_DGRTT


def get_membres():
    """Retourne la liste des membres de la DR&I."""
    from labster.domain.models.profiles import Profile

    return Profile.query.filter(Profile.dgrtt == True).order_by(Profile.nom).all()


def est_referent(user):
    from labster.domain.models.mapping_dgrtt import MappingDgrtt

    return (
        MappingDgrtt.query.filter_by(contact_dgrtt=user)
        .filter_by(bureau_dgrtt="REF")
        .count()
        > 0
    )


def mapping():
    from labster.domain.models.mapping_dgrtt import MappingDgrtt

    result = {}
    mappings = MappingDgrtt.query.options(
        joinedload(MappingDgrtt.contact_dgrtt), joinedload(MappingDgrtt.ou_recherche)
    ).all()

    structures = {m.ou_recherche for m in mappings}
    structures = sorted(structures)
    for structure in structures:
        line = []
        for bureau in bureaux():
            contacts = [
                m.contact_dgrtt
                for m in mappings
                if m.bureau_dgrtt == bureau.id and m.ou_recherche == structure
            ]
            if contacts:
                contact = contacts[0]
            else:
                contact = None
            line.append(contact)

        result[structure] = line
    return result


def contacts_structure(structure):
    """Retourne un dictionnaire 'sigle bureau' -> contact pour une structure
    donnée."""
    from labster.domain.models.mapping_dgrtt import MappingDgrtt

    result = {}
    for bureau in bureaux():
        mappings = (
            MappingDgrtt.query.filter(MappingDgrtt.ou_recherche == structure)
            .filter(MappingDgrtt.bureau_dgrtt == bureau.id)
            .all()
        )
        assert len(mappings) in [0, 1]
        if mappings:
            result[bureau.id] = mappings[0].contact_dgrtt
        else:
            result[bureau.id] = None
    return result


def set_mapping(structure, bureau_id, contact):
    from labster.domain.models.mapping_dgrtt import MappingDgrtt

    mappings_to_delete = (
        MappingDgrtt.query.filter(MappingDgrtt.ou_recherche == structure)
        .filter(MappingDgrtt.bureau_dgrtt == bureau_id)
        .all()
    )
    assert len(mappings_to_delete) in [0, 1]
    for m in mappings_to_delete:
        db.session.delete(m)

    mapping = MappingDgrtt(
        ou_recherche=structure, bureau_dgrtt=bureau_id, contact_dgrtt=contact
    )
    db.session.add(mapping)


def get_referent(structure):
    """Retourne le référent de la structure, ou None s'il n'y en a pas."""
    from labster.domain.models.mapping_dgrtt import MappingDgrtt

    mappings = (
        MappingDgrtt.query.filter_by(bureau_dgrtt="REF")
        .filter_by(ou_recherche=structure)
        .all()
    )
    assert len(mappings) in (0, 1)

    if len(mappings) == 0:
        return None
    return mappings[0].contact_dgrtt


def labos_dont_je_suis_referent(user):
    """Retourne la liste des structure dont l'utilisateur est référent."""
    from labster.domain.models.mapping_dgrtt import MappingDgrtt

    mappings = (
        MappingDgrtt.query.filter_by(bureau_dgrtt="REF")
        .filter_by(contact_dgrtt=user)
        .all()
    )

    return {m.ou_recherche for m in mappings}


def get_contact_dgrtt(structure, bureau_dgrtt):
    # type: (OrgUnit, Text) -> Optional[Profile]
    assert structure and bureau_dgrtt
    assert isinstance(bureau_dgrtt, str)
    from labster.domain.models.mapping_dgrtt import MappingDgrtt

    mappings = (
        MappingDgrtt.query.filter_by(ou_recherche=structure)
        .filter_by(bureau_dgrtt=bureau_dgrtt)
        .all()
    )

    assert len(mappings) in [0, 1]
    if not mappings:
        return None
    return mappings[0].contact_dgrtt


def get_membres_du_bureau_dgrtt(bureau_dgrtt: BureauDgrtt) -> list[Profile]:
    """Retourne la liste des membres d'un bureau DR&I (DGRTT) donné."""
    from labster.domain.models.mapping_dgrtt import MappingDgrtt

    if not bureau_dgrtt:
        return []

    assert isinstance(bureau_dgrtt, BureauDgrtt)

    mappings = MappingDgrtt.query.filter(
        MappingDgrtt.bureau_dgrtt == bureau_dgrtt.id
    ).all()

    users = [m.contact_dgrtt for m in mappings]
    users = sorted(users, key=lambda x: x.uid)
    return users


def get_bureau_dgrtt(user, check=False):
    # type: (Profile, bool) -> Optional[BureauDgrtt]
    """Retourne le bureau DR&I (DGRTT) d'appartenance d'un contact DGRTT."""
    if user.chef_du_bureau:
        return BureauDgrtt.from_id(user.chef_du_bureau)

    from labster.domain.models.mapping_dgrtt import MappingDgrtt

    mappings = MappingDgrtt.query.filter(MappingDgrtt.contact_dgrtt == user).all()

    bureau_ids = {m.bureau_dgrtt for m in mappings if m.bureau_dgrtt != "REF"}
    if len(bureau_ids) > 1:
        flash(
            f"L'utilisateur {user.full_name} appartient à plus d'un bureau",
            category="warning",
        )
        if check:
            raise ValueError(
                f"L'utilisateur {user.full_name} appartient à plus d'un bureau"
            )
    if not bureau_ids:
        return None
    return BureauDgrtt.from_id(bureau_ids.pop())


def get_perimetre_dgrtt(user):
    # type: (Profile) -> Set[OrgUnit]
    from labster.domain.models.mapping_dgrtt import MappingDgrtt

    mappings = MappingDgrtt.query.filter(MappingDgrtt.contact_dgrtt == user).all()

    return {m.ou_recherche for m in mappings}


def check():
    membres_dgrtt = get_membres()
    for membre in membres_dgrtt:
        perimetre = get_perimetre_dgrtt(membre)
        for structure in perimetre:
            contacts = contacts_structure(structure).values()
            assert membre in contacts

    from labster.domain.models.unites import OrgUnit

    for structure in OrgUnit.query.all():
        contacts = contacts_structure(structure)
        for sigle_bureau, contact in contacts.items():
            if contact and sigle_bureau != "REF":
                bureau = get_bureau_dgrtt(contact, check=True)
                assert bureau.id == sigle_bureau

        referent = get_referent(structure)
        if referent:
            assert referent.has_role("référent")
            assert structure in labos_dont_je_suis_referent(referent)



from __future__ import annotations

from functools import singledispatch
from pathlib import Path

from flask import Flask
from whoosh import index
from whoosh.analysis import StemmingAnalyzer
from whoosh.fields import ID, KEYWORD, TEXT, Schema

from labster.types import JSONDict

schema = Schema(
    key=ID(unique=True),
    id=ID(stored=True),
    cls=KEYWORD(stored=True),
    name=TEXT(stored=True),
    text=TEXT(analyzer=StemmingAnalyzer()),
)


class Whoosh:
    whoosh_base: str
    app: Flask | None

    def __init__(self, app=None):
        self.app = app
        if app is not None:
            self.init_app(app)

    def init_app(self, app: Flask) -> None:
        whoosh_base = Path(app.config.get("WHOOSH_BASE", "whoosh"))

        if not whoosh_base.is_absolute():
            whoosh_base = Path(app.instance_path) / whoosh_base

        if not whoosh_base.is_dir():
            whoosh_base.mkdir(parents=True)

        self.whoosh_base = str(whoosh_base.resolve())

        if index.exists_in(self.whoosh_base):
            self.index = index.open_dir(self.whoosh_base)
        else:
            self.index = index.create_in(self.whoosh_base, schema)

    def index_object(self, obj, writer=None):
        from labster.domain2.model.demande import Demande

        @singledispatch
        def as_document(obj):
            raise NotImplementedError

        @as_document.register(Demande)
        def as_document_demande(demande: Demande) -> JSONDict:
            kw: JSONDict = {}

            indexed_values = []
            for k in ["nom", "name", "no_infolab", "no_eotp"]:
                v = getattr(demande, k)
                if v:
                    indexed_values.append(v)

            for v in demande.data.values():
                if isinstance(v, str):
                    indexed_values.append(v)

            allowed_roles_and_users: list[str] = []

            v = demande.porteur
            if v:
                indexed_values.append(v.full_name)
                allowed_roles_and_users += [f"user:{v.id}"]

            v = demande.gestionnaire
            if v:
                indexed_values.append(v.full_name)
                allowed_roles_and_users += [f"user:{v.id}"]

            # v = obj.laboratoire
            # if v:
            #     indexed_values.append(v.nom)
            #
            # v = obj.structure
            # if v:
            #     indexed_values.append(v.nom)
            #     allowed_roles_and_users += [f"org:{v.id}"]

            kw["id"] = str(demande.id)
            kw["text"] = " ".join(indexed_values)
            kw["name"] = demande.nom
            kw["cls"] = "Demande"
            kw["key"] = f"demande:{demande.id}"

            # kw["allowed_roles_and_users"] = " ".join(allowed_roles_and_users)
            return kw

        must_commit = False
        if not writer:
            writer = self.index.writer()
            must_commit = True

        document = as_document(obj)
        writer.update_document(**document)

        if must_commit:
            writer.commit()



from __future__ import annotations

from datetime import date, datetime
from decimal import Decimal
from enum import Enum, unique
from functools import singledispatch
from typing import Any, Collection

import dateutil
import sqlalchemy as sa
from abilian.app import db
from email_validator import EmailNotValidError, validate_email
from sqlalchemy import JSON, Boolean, Column, Date, DateTime, ForeignKey, \
    Integer, String
from sqlalchemy.ext.mutable import MutableDict, MutableList
from sqlalchemy.orm import relationship

from labster.domain2.model.base import Repository
from labster.domain2.model.profile import Profile
from labster.domain2.model.structure import Structure, StructureRepository
from labster.domain2.model.util import parse_date
from labster.domain2.services.calculs_couts import cout_total_charge
from labster.domain2.services.contacts import ContactService, ContactType
from labster.domain2.services.workflow.states import EN_EDITION
from labster.domain2.services.workflow.workflow import LabsterWorkflow
from labster.lib.workflow import State


@unique
class DemandeType(Enum):
    CONVENTION = "Convention de recherche"
    AVENANT_CONVENTION = "Avenant à une convention de recherche"
    RECRUTEMENT = "Recrutement"
    PI_LOGICIEL = "Déclaration de logiciel / base de données"
    PI_INVENTION = "Déclaration d´invention"
    AUTRE = "Demande autre"
    # To please the typechecker
    ILLEGAL = "Should not happen"


types_demande = [x.value for x in DemandeType]


class Demande(db.Model):
    __tablename__ = "v3_demandes"

    _type: str = DemandeType.ILLEGAL.value

    id = Column(Integer, primary_key=True)
    old_id = Column(Integer)
    type = Column(
        sa.Enum(*types_demande, name="type_demande"),
        nullable=False,
        index=True,
    )
    created_at = Column(DateTime)
    updated_at = Column(DateTime)
    #
    nom = Column(String)
    name = Column(String)
    active = Column(Boolean, default=True)
    editable = Column(Boolean, default=True)
    no_infolab = Column(String)
    no_eotp = Column(String)

    # Relations
    contact_labco_id = Column(String(36), ForeignKey(Profile.id))
    gestionnaire_id = Column(String(36), ForeignKey(Profile.id))
    porteur_id = Column(String(36), ForeignKey(Profile.id))
    structure_id = Column(String(36), ForeignKey(Structure.id))

    #
    data = Column(MutableDict.as_mutable(JSON))
    past_versions = Column(MutableList.as_mutable(JSON))
    form_state = Column(MutableDict.as_mutable(JSON))
    attachments = Column(MutableDict.as_mutable(JSON))
    feuille_cout = Column(MutableDict.as_mutable(JSON))
    documents_generes = Column(MutableList.as_mutable(JSON))

    # Workflow
    # wf_state = Column(WF_ENUM, default=EN_EDITION.id, nullable=False, index=True)
    wf_state = Column(String)
    wf_date_derniere_action = Column(DateTime)
    wf_retard = Column(Integer)
    wf_history = Column(MutableList.as_mutable(JSON))
    wf_data = Column(MutableDict.as_mutable(JSON))
    #: Date de validation par la hiérarchie
    date_effective = Column(Date, nullable=True)

    contact_labco = relationship(
        Profile, foreign_keys=lambda: [Demande.contact_labco_id]
    )
    porteur = relationship(Profile, foreign_keys=lambda: [Demande.porteur_id])
    gestionnaire = relationship(Profile, foreign_keys=lambda: [Demande.gestionnaire_id])
    structure = relationship(Structure)

    __mapper_args__ = {"polymorphic_identity": "", "polymorphic_on": type}

    # wf_stage_id = Column(Integer, ForeignKey(OrgUnit.id), index=True, nullable=True)
    # wf_stage = relationship(
    #     OrgUnit, primaryjoin=remote(Entity.id) == foreign(wf_stage_id)
    # )
    #
    # #: id de la personne responsable de la tâche en cours
    # wf_current_owner_id = Column(
    #     Integer, ForeignKey(Profile.id), index=True, nullable=True
    # )
    # #: la personne responsable de la tâche en cours
    # wf_current_owner = relationship(
    #     Profile, primaryjoin=remote(Entity.id) == foreign(wf_current_owner_id)
    # )

    # id: Optional[DemandeId] = None
    # created_at: datetime = field(default_factory=datetime.utcnow)
    #
    # nom: str = ""
    # name: str = ""
    # no_infolab: str = ""
    # no_eotp: str = ""
    #
    # # type = Column(Enum(*TYPE_ENUM, name="demande_type"), nullable=False, index=True)
    #
    # data: JSONDict = field(default_factory=dict)
    #
    # past_versions: JSONList = field(default_factory=list)
    # form_state: JSONDict = field(default_factory=dict)
    # attachments: JSONDict = field(default_factory=dict)
    # feuille_cout: JSONDict = field(default_factory=dict)
    # documents_generes: JSONList = field(default_factory=list)
    #
    # #: Date de validation par la hiérarchie
    # date_effective: Optional[date] = None
    #
    # #: Seules les demandes actives apparaissent dans le workflow.
    # #: Les autres sont considérées comme archivées.
    # active: bool = field(default_factory=lambda: True)
    # editable: bool = field(default_factory=lambda: True)
    #
    # # Les acteurs de la demande:
    # porteur: Optional[Profile] = None
    # gestionnaire: Optional[Profile] = None
    # contact_labco: Optional[Profile] = None
    #
    # # Structures liées
    # structure: Optional[Structure] = None
    #
    # # Variables liées au workflow [TODO]
    # wf_state: str = EN_EDITION.id
    # wf_date_derniere_action: Optional[datetime] = None
    # wf_retard: int = 0
    # wf_history: List[Dict[str, Any]] = field(default_factory=list)
    # wf_data: JSONDict = field(default_factory=dict)
    # wf_current_owner: Optional[Profile] = None

    # TODO:
    # wf_stage: a remplacer

    # wf_state = Column(WF_ENUM, default=EN_EDITION.id, nullable=False, index=True)
    #
    # wf_date_derniere_action = Column(DateTime, nullable=False)
    #
    # #: nombre de jours de retard (0 si pas de retard)
    # wf_retard = Column(Integer, nullable=False, default=0)
    #
    # wf_history = Column(JSONList(), default=list)
    #
    # wf_data = Column(JSONDict(), default=dict)
    #
    # wf_stage_id = Column(Integer, ForeignKey(OrgUnit.id), index=True, nullable=True)
    # wf_stage = relationship(
    #     OrgUnit, primaryjoin=remote(Entity.id) == foreign(wf_stage_id)
    # )
    #
    # #: id de la personne responsable de la tâche en cours
    # wf_current_owner_id = Column(
    #     Integer, ForeignKey(Profile.id), index=True, nullable=True
    # )
    # #: la personne responsable de la tâche en cours
    # wf_current_owner = relationship(
    #     Profile, primaryjoin=remote(Entity.id) == foreign(wf_current_owner_id)
    # )
    #
    # __mapper_args__ = {"polymorphic_identity": "", "polymorphic_on": type}

    def __init__(self, **kw):
        # if not hasattr(self, "porteur"):
        #     self.porteur = None
        # if not hasattr(self, "gestionnaire"):
        #     self.gestionnaire = None

        # assert self.porteur or self.gestionnaire
        # if self.porteur:
        #     self.structure = self.porteur.structure

        self.data = {}
        self.attachments = {}
        self.form_state = {"fields": []}
        self.versions = []
        self.wf_state = EN_EDITION.id
        self.wf_history = []
        self.wf_data = {}
        self.wf_retard = 0
        self.feuille_cout = {}
        self.editable = True
        self.active = True
        if not self.created_at:
            self.created_at = datetime.utcnow()
        self.wf_date_derniere_action = self.created_at

        super().__init__(**kw)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} with id={self.id}>"

    def log_creation(self, actor: Profile) -> None:
        if not self.wf_history:
            message = f"Demande créée par l'utilisateur {actor.full_name}"
            log_entry = {
                "date": datetime.now().strftime("%d %b %Y %H:%M:%S"),
                "actor_id": actor.id,
                "message": message,
                "note": "",
            }
            self.wf_history = [log_entry]

    def clone(self) -> Demande:
        return self.__class__(
            nom=self.nom,
            wf_state=EN_EDITION.id,
            porteur=self.porteur,
            gestionnaire=self.gestionnaire,
            structure=self.structure,
            data=self.data.copy(),
            form_state=self.form_state.copy(),
        )

    @property
    def date_debut(self):
        return None

    @property
    def age(self):
        if not self.date_effective:
            return 0
        dt = date.today() - self.date_effective
        return int(dt.days)

    @property
    def retard(self):
        if not self.wf_date_derniere_action or not self.active:
            return 0

        # TODO: implémenter la logique de jours ouvrés
        dt = datetime.utcnow() - self.wf_date_derniere_action
        return int(dt.days)

    def update_retard(self):
        # TODO: implémenter la logique de jours ouvrés
        dt = datetime.utcnow() - self.wf_date_derniere_action
        self.wf_retard = int(dt.days)

    def nom_par_defaut(self) -> str:
        return self.type + " sans nom"

    #
    # Accessors / properties
    #
    def __getattr__(self, name: str) -> Any:
        if name.startswith("_"):
            raise AttributeError(f"object has no attribute '{name}'")

        data = getattr(self, "data", None)
        if data and name in data:
            return data[name]

        raise AttributeError(f"object has no attribute '{name}'")

    def has_same_data(self, data: dict[str, Any]) -> bool:
        """Return True if the current version is the same as the given data."""
        current_data = self.data

        for k in set(current_data.keys()) | set(data.keys()):
            if k in ["csrf_token"]:
                continue
            if k.startswith("html-"):
                continue
            current_value = current_data.get(k)
            new_value = data.get(k)
            if current_value != new_value:
                return False
        return True

    @property
    def contact(self) -> Profile:
        result = self.gestionnaire or self.porteur
        assert result
        return result

    @property
    def directeur_name(self) -> str:
        structure = self.structure
        if structure and structure.directeur:
            return structure.directeur.full_name
        return ""

    @property
    def owners(self) -> set[Profile]:
        owners = set()
        if self.gestionnaire:
            owners.add(self.gestionnaire)
        if self.porteur:
            owners.add(self.porteur)
        return owners

    #
    # Workflow
    #
    def get_workflow(self, user: Profile | None = None) -> LabsterWorkflow:
        return LabsterWorkflow(self, user)

    def get_state(self, user: Profile | None = None) -> State:
        workflow = self.get_workflow(user)
        return workflow.current_state()

    def current_owners(self) -> list[Profile]:
        return self.get_workflow().current_owners()

    def structures_signataires(self) -> set[Structure]:
        result = self.get_structure_concernees()
        if self.structure:
            result.add(self.structure)
        return result

    def valideurs(self) -> set[Profile]:
        structures_signataires = self.structures_signataires()
        result = set()
        for structure in structures_signataires:
            result.update(structure.responsables)
        return result

    def get_structure_concernees(self):
        from labster.di import injector

        structure_repo = injector.get(StructureRepository)

        result = set()
        if self.data.get("structures_concernees"):
            for d in self.structures_concernees:
                structure_id = d["value"]
                structure = structure_repo.get_by_id(structure_id)
                if structure:
                    result.add(structure)

        return result

    @property
    def date_soumission(self) -> date | None:
        for entry in self.wf_history:
            if entry.get("transition") == "SOUMETTRE":
                return dateutil.parser.parse(entry["date"]).date()

        return None

    @property
    def date_finalisation(self) -> date | None:
        final_states = ["CONFIRMER_FINALISATION_DGRTT", "ABANDONNER", "REJETER_DGRTT"]
        for entry in self.wf_history:
            if entry.get("transition") in final_states:
                return dateutil.parser.parse(entry["date"]).date()

        return None

    def assigne_contact_labco(self):
        if self.contact_labco:
            return

        bureau = get_bureau_dri(self)
        if bureau:
            self.contact_labco = get_contact_labco(self.structure, bureau)
        else:
            # TODO: special case: demande "Autre"
            pass

    #
    # Data validation / manipulation
    #
    def validate(self) -> Validation:
        return Validation(self, self.get_errors(), self.get_extra_errors())

    def get_errors(self) -> list[Any]:
        errors = []
        form_state = self.form_state
        fields = form_state.get("fields", [])
        for field_name, field_value in self.data.items():
            if field_name not in fields:
                continue
            field: dict[str, Any] = fields[field_name]

            visible = bool(field.get("visible"))
            required = bool(field.get("required"))

            if visible and required and not field_value:
                errors.append(field["name"])

            if visible and "mail" in field_name and field_value:
                try:
                    validate_email(field_value)
                except EmailNotValidError:
                    errors.append(field_name)

        return errors

    def get_extra_errors(self) -> list[str]:
        return []

    def is_valid(self) -> bool:
        validation = self.validate()
        return validation.ok

    @property
    def errors(self):
        return self.validate().errors

    def update_data(self, data: dict) -> None:
        self.increase_version()
        self.data.update(data)

        self.update_nom()
        self.post_update()

    # new_name = data.get('nom', None)
    # if new_name is not None and new_name != self.name:
    #     self.name = new_name

    def post_update(self) -> None:
        new_porteur_uid = self.data.get("porteur", None)

        if new_porteur_uid is not None:
            try:
                new_porteur = db.session.query(Profile).get_by_uid(new_porteur_uid)
                self.porteur = new_porteur
                self.structure = new_porteur.structure
            except Exception:
                # TODO: better solution to deal with tests.
                pass

    def increase_version(self) -> None:
        if self.data:
            if not self.past_versions:
                self.past_versions = []
            self.past_versions.append(
                (self.data, datetime.utcnow().strftime("%d %b %Y %H:%M:%S"))
            )

    # Pièces jointes: TODO remove or debug.
    # @property
    # def pieces_jointes(self) -> List[Dict]:
    #     result = []
    #     for k, v in sorted(self.attachments.items()):
    #         if isinstance(v, dict):
    #             creator = db.session.query(Profile).filter(Profile.login==v["creator"]).first()
    #
    #             d = {"id": v["id"], "name": k, "creator": creator}
    #
    #             date_str = v.get("date")
    #             if date_str:
    #                 d["date"] = iso8601.parse_date(date_str)
    #             else:
    #                 d["date"] = None
    #         else:
    #             d = {"id": v, "name": k, "creator": None, "date": None}
    #         id = d["id"]
    #         blob = Blob.query.get(id)
    #         if blob:
    #             result.append(d)
    #     return result


#
# Concrete classes
#
class DemandeRH(Demande):
    _type = DemandeType.RECRUTEMENT.value
    icon_class = "far fa-user"

    __mapper_args__ = {"polymorphic_identity": _type}

    def update_nom(self) -> None:
        prenom = self.data.get("prenom") or "(prénom inconnu)"
        nom = self.data.get("nom") or "(nom inconnu)"
        nature = (
            self.data.get("nature_du_recrutement") or "(nature du recrutement inconnue)"
        )
        self.nom = f"Recrutement de {prenom} {nom} - {nature}"
        self.name = self.nom

    @property
    def date_debut(self) -> date | None:
        return parse_date(self.data.get("date_debut"))

    @property
    def date_fin(self) -> date | None:
        return parse_date(self.data.get("date_fin"))

    @property
    def duree_jours(self) -> int:
        if not (self.date_fin and self.date_debut):
            return 0
        return (self.date_fin - self.date_debut).days + 1

    @property
    def duree_mois(self) -> int:
        if not (self.date_fin and self.date_debut):
            return 0
        years = self.date_fin.year - self.date_debut.year
        months = self.date_fin.month - self.date_debut.month
        days = self.date_fin.day - self.date_debut.day
        return 12 * years + months + round(days / 30.4375)
        # return (self.date_fin - self.date_debut).days + 1

    @property
    def cout_total_charge(self) -> Decimal:
        try:
            return cout_total_charge(self)
        except Exception:
            return Decimal(0)

    def get_extra_errors(self) -> list[str]:
        errors: list[str] = []
        if not self.attachments:
            errors += ["Vous devez obligatoirement attacher une pièce-jointe."]
        if self.date_fin and self.date_debut and self.date_fin <= self.date_debut:
            errors += ["La date de fin est antérieure à la date de début."]
        return errors


class DemandeConvention(Demande):
    _type = DemandeType.CONVENTION.value
    icon_class = "far fa-briefcase"

    __mapper_args__ = {"polymorphic_identity": _type}

    @property
    def date_debut(self):
        return parse_date(self.data.get("date_depot"))

    def update_nom(self):
        nom_ou_acronyme = self.data.get("nom_ou_acronyme")
        if nom_ou_acronyme:
            self.nom = nom_ou_acronyme
        else:
            self.nom = self.nom_par_defaut()

    def get_extra_errors(self):
        errors = []
        if self.appel_a_projets == "oui":
            if not self.attachments:
                errors += [
                    "Vous devez renseigner la liste des partenaires en pièce-jointe."
                ]

        if self.appel_a_projets == "non" and not self.partenaires:
            errors += ["Vous devez saisir au moins un partenaire dans le formulaire."]

        return errors


class DemandeAvenantConvention(Demande):
    _type = DemandeType.AVENANT_CONVENTION.value
    icon_class = "far fa-briefcase"

    __mapper_args__ = {"polymorphic_identity": _type}

    def update_nom(self):
        nom_ou_acronyme = self.data.get("nom_projet")
        if nom_ou_acronyme:
            self.nom = nom_ou_acronyme
        else:
            self.nom = self.nom_par_defaut()

    @property
    def modifications(self):
        modifs = []
        if self.montant == "oui":
            modifs.append("montant")
        if self.duree == "oui":
            modifs.append("durée")
        if self.programme_scientifique == "oui":
            modifs.append("programme scientifique")
        if self.consortium == "oui":
            modifs.append("consortium")
        if self.autre == "oui":
            modifs.append("autre")
        return ", ".join(modifs) + "."


# Abstract
class DemandePiMixin:
    intitule = ""
    acronyme = ""
    titre = ""

    def update_nom(self):
        intitule = self.intitule or self.titre
        acronyme = self.acronyme
        if acronyme and intitule:
            self.nom = f"{intitule} ({acronyme})"
        elif acronyme or intitule:
            self.nom = acronyme or intitule
        else:
            self.nom = self.nom_par_defaut()


# Concrete
class DemandePiLogiciel(Demande, DemandePiMixin):
    _type = DemandeType.PI_LOGICIEL.value
    icon_class = "far fa-save"

    __mapper_args__ = {"polymorphic_identity": _type}

    @property
    def intitule(self):
        return self.data.get("intitule", "")

    @property
    def acronyme(self):
        return self.data.get("acronyme", "")


class DemandePiInvention(Demande, DemandePiMixin):
    _type = DemandeType.PI_INVENTION.value
    icon_class = "far fa-rocket"

    __mapper_args__ = {"polymorphic_identity": _type}

    @property
    def titre(self):
        return self.data.get("titre", "")


class DemandeAutre(Demande):
    _type = DemandeType.AUTRE.value
    commentaire = ""
    icon_class = "far fa-folder-open"

    __mapper_args__ = {"polymorphic_identity": _type}

    @property
    def titre(self):
        return self.data.get("titre", "")

    def update_nom(self):
        self.nom = self.titre


_REGISTRY: dict[str, type[Demande]] = {
    "rh": DemandeRH,
    "convention": DemandeConvention,
    "avenant_convention": DemandeAvenantConvention,
    "pi_logiciel": DemandePiLogiciel,
    "pi_invention": DemandePiInvention,
    "autre": DemandeAutre,
}


def demande_factory(type: str, demandeur: Profile, data: dict, **args: dict) -> Demande:
    demande_cls = _REGISTRY.get(type)
    if not demande_cls:
        raise RuntimeError(f"Type de demande illégal: {type}")

    demande = demande_cls(**args)
    demande.update_data(data)
    demande.log_creation(demandeur)
    return demande


class Validation:
    def __init__(
        self, obj: Demande, errors: list[Any], extra_errors: list[str]
    ) -> None:
        self.obj = obj
        self.errors = errors
        self.extra_errors = extra_errors

    @property
    def ok(self) -> bool:
        return (not self.errors) and (not self.extra_errors)


class DemandeRepository(Repository):
    def get_all(self) -> Collection[Demande]:
        raise NotImplementedError


@singledispatch
def get_bureau_dri(demande) -> ContactType:
    raise NotImplementedError


@get_bureau_dri.register
def get_bureau_dri_convention(demande: DemandeConvention):
    return _get_bureau_dri_convention(demande)


@get_bureau_dri.register
def get_bureau_dri_avenant_convention(demande: DemandeAvenantConvention):
    return _get_bureau_dri_convention(demande)


def _get_bureau_dri_convention(demande: Demande):
    """Le choix dans cette liste et la case à cocher "Le projet intègre-t-
    il une entreprise ? " pour ANR détermine le destinataire des infos du
    formulaire:

    - Si "ANR" est choisi et "non " coché = bureau des contrats publics
    - Si "ANR" est choisi et "oui " coché = bureau Entreprises
    - Si "Commission européenne" est choisi = bureau Europe
    - Si "Entreprise" est choisi = bureau Entreprises
    - Si "Financement public national" est choisi = bureau des contrats publics
    - Si "FUI" est choisi  = bureau Entreprises
    - Si "Institution européenne et internationale" est choisi = bureau des contrats publics
    - Si "Autre" est choisi  et "oui " coché = bureau Entreprises
    - Si "Autre" est choisi  et "non " coché = référent du laboratoire

    +

    Champ Type de financeur : il y a un problème d'adressage du formulaire
    quand Type de financeur=ANR.
    Si Type de financeur=ANR ET "Le projet intègre-t-il une entreprise=non"
    le formulaire est à envoyer au bureau Contrats publics
    Mais si Type de financeur=ANR ET "Le projet intègre-t-il une entreprise=oui",
    le formulaire est à envoyer au bureau Entreprises et transfert de technologie
    """

    type_financeur = demande.data.get("type_financeur")
    if not type_financeur:
        return

    integre_entreprise = demande.data.get("integre_entreprise")
    if integre_entreprise == "non":
        integre_entreprise = False

    bureau = None
    if type_financeur == "ANR":
        if integre_entreprise:
            bureau = ContactType.CONTACT_ENTREPRISES
        else:
            bureau = ContactType.CONTACT_CONTRATS_PUBLICS
    elif type_financeur == "Commission Européenne":
        bureau = ContactType.CONTACT_EUROPE
    elif type_financeur == "Entreprise":
        bureau = ContactType.CONTACT_ENTREPRISES
    elif type_financeur == "Financement public national":
        bureau = ContactType.CONTACT_CONTRATS_PUBLICS
    elif type_financeur == "FUI":
        bureau = ContactType.CONTACT_ENTREPRISES
    elif type_financeur == "Institution européenne et internationale":
        bureau = ContactType.CONTACT_CONTRATS_PUBLICS
    elif type_financeur == "Fondation":
        bureau = ContactType.CONTACT_CONTRATS_PUBLICS
    elif type_financeur == "Autre":
        if integre_entreprise:
            bureau = ContactType.CONTACT_ENTREPRISES
        else:
            bureau = ContactType.CDP

    assert bureau
    return bureau


@get_bureau_dri.register
def get_bureau_dri_pi(demande: DemandePiMixin):
    return ContactType.CONTACT_PI


@get_bureau_dri.register
def get_bureau_dri_rh(demande: DemandeRH):
    return ContactType.CONTACT_RH


@get_bureau_dri.register
def get_bureau_dri_autre(demande: DemandeAutre):
    return ContactType.CONTACT_DRV


def get_contact_labco(structure: Structure, type: ContactType) -> Profile | None:
    from labster.di import injector

    contact_service = injector.get(ContactService)
    return contact_service.get_contact(structure, type)



from __future__ import annotations

from flask_sqlalchemy import SQLAlchemy
from injector import inject
from sqlalchemy.orm import Session
from sqlalchemy.orm.exc import NoResultFound

from labster.domain2.model.profile import Profile, ProfileId, ProfileRepository
from labster.infrastructure.repositories.sqla.mappers import Mapper


class SqlaProfileRepository(ProfileRepository):
    session: Session

    @inject
    def __init__(self, db: SQLAlchemy, mapper: Mapper):
        self.db = db
        self.session = db.session

    def query(self):
        return self.session.query(Profile)

    def get_all(self) -> set[Profile]:
        return set(self.query().all())

    def put(self, profile: Profile):
        if not profile.id:
            profile.id = ProfileId.new()
        self.session.add(profile)

    def delete(self, profile: Profile):
        self.session.delete(profile)

    def is_empty(self):
        return self.query().count() == 0

    def clear(self):
        self.query().delete()

    def get_by_id(self, id: ProfileId) -> Profile:
        result = self.query().get(id)
        if not result:
            raise NoResultFound
        return result

    def get_by_old_id(self, old_id: int) -> Profile:
        return self.query().filter_by(old_id=old_id).one()

    def get_by_login(self, login: str) -> Profile:
        return self.query().filter_by(login=login).one()

    def get_by_uid(self, uid: str) -> Profile:
        return self.query().filter_by(uid=uid).one()

    def get_by_old_uid(self, old_uid: str) -> Profile:
        return self.query().filter_by(old_uid=old_uid).one()



from __future__ import annotations

import uuid
from collections import defaultdict
from dataclasses import dataclass
from typing import Any

from flask_sqlalchemy import SQLAlchemy
from injector import inject
from sqlalchemy import Column, ForeignKey, String, Table
from sqlalchemy.orm import joinedload, mapper, relationship

from labster.domain2.model.profile import Profile
from labster.domain2.model.structure import Structure
from labster.domain2.services.roles import Role, RoleService


def new_id() -> str:
    return str(uuid.uuid4())


@dataclass
class Grant:
    id: str
    user_id: str
    role_name: str
    context_id: str | None


def make_mapper(metadata):
    table = Table(
        "v3_grants",
        metadata,
        #
        Column("id", String(36), primary_key=True),
        Column("user_id", String(36), ForeignKey(Profile.id), index=True),
        Column("role_name", String(64), index=True, nullable=False),
        Column("context_id", String(36), ForeignKey(Structure.id), index=True),
    )

    mapper(
        Grant,
        table,
        properties={"context": relationship(Structure), "user": relationship(Profile)},
    )


class SqlaRoleService(RoleService):
    @inject
    def __init__(
        self,
        db: SQLAlchemy,
    ):
        self.db = db
        self.session = self.db.session
        make_mapper(self.db.metadata)

    def query(self):
        return self.session.query(Grant)

    def is_empty(self):
        return self.query().count() == 0

    def clear(self):
        self.query().delete()

    def grant_role(self, user: Profile, role: Role, context: Any = None):
        grant = self.get_grant(user, role, context)
        if grant:
            return

        if context:
            grant = Grant(new_id(), user.id, role.name, context.id)
        else:
            grant = Grant(new_id(), user.id, role.name, None)

        self.session.add(grant)
        self.session.flush()

    def ungrant_role(self, user: Profile, role: Role, context: Any = None):
        grant = self.get_grant(user, role, context)
        if grant:
            self.session.delete(grant)
            self.session.flush()

    def get_grant(self, user: Profile, role: Role, context) -> Grant:
        assert isinstance(user, Profile)
        assert isinstance(role, Role)

        if context:
            context_id = context.id
        else:
            context_id = None
        query = (
            self.query()
            .filter_by(user_id=user.id)
            .filter_by(role_name=role.name)
            .filter_by(context_id=context_id)
        )
        return query.first()

    def has_role(self, user: Profile, role: Role, context: Any = None):
        assert context in ("*", None) or isinstance(context, Structure), str(context)

        if context == "*":
            query = (
                self.query().filter_by(user_id=user.id).filter_by(role_name=role.name)
            )
            return query.first() is not None
        else:
            return self.get_grant(user, role, context) is not None

    def get_users_with_role(self, role: Role, context: Any = None) -> set[Profile]:
        assert isinstance(role, Role)

        query = self.query().filter_by(role_name=role.name)
        if isinstance(context, Structure):
            query = query.filter_by(context_id=context.id)
        elif context is None:
            query = query.filter_by(context_id=None)
        elif context == "*":
            pass
        else:
            raise TypeError(f"Illegal argument for context: {context}")

        grants = query.options(joinedload(Grant.user)).all()
        return {grant.user for grant in grants}

    def get_users_with_role_on(self, context: Structure) -> dict[Role, set[Profile]]:
        query = self.query().filter_by(context_id=context.id)
        grants = query.all()

        result: dict[Role, set] = defaultdict(set)
        for grant in grants:
            role = getattr(Role, grant.role_name)
            result[role].add(grant.user)

        return result

    def get_users_with_given_role(self, role: Role, context: Structure) -> set[Profile]:
        assert isinstance(role, Role)
        assert isinstance(context, Structure)

        query = self.query().filter(Grant.context_id == context.id)
        if role == Role.MEMBRE:
            query = query.filter(
                Grant.role_name.in_(
                    [
                        Role.MEMBRE.name,
                        Role.MEMBRE_AFFILIE.name,
                        Role.MEMBRE_AFFECTE.name,
                        Role.MEMBRE_RATTACHE.name,
                    ]
                )
            )
        else:
            query = query.filter(Grant.role_name == role.name)
        grants = query.all()
        return {grant.user for grant in grants}

    def get_roles_for_user(self, user: Profile) -> dict[Role, set[Structure]]:
        query = self.query().filter_by(user_id=user.id)
        grants = query.all()

        result = defaultdict(set)
        for grant in grants:
            role = getattr(Role, grant.role_name)
            structure = grant.context
            if structure and not structure.active:
                continue
            result[role].add(structure)

        return result



"""Modèle de case management adapté au projet Labster."""
from __future__ import annotations

from datetime import date, datetime
from decimal import Decimal
from typing import TYPE_CHECKING, Any, Collection

import dateutil.parser
import structlog
from abilian.core.entities import Entity
from abilian.core.models import IdMixin, TimestampedMixin
from abilian.core.models.base import SEARCHABLE, Indexable
from abilian.core.models.blob import Blob
from abilian.core.sqlalchemy import JSONDict, JSONList
from flask_sqlalchemy import BaseQuery
from iso8601 import iso8601
from sqlalchemy import Boolean, Column, Date, DateTime, Enum, ForeignKey, \
    Integer, Unicode
from sqlalchemy.orm import foreign, relationship, remote

from labster.domain.models.util import parse_date
from labster.domain.services.calculs_couts import cout_total_charge
from labster.domain.services.dgrtt import get_contact_dgrtt
from labster.extensions import db

from .profiles import Profile
from .unites import OrgUnit
from .workflow import ALL_STATES, EN_EDITION, LabsterWorkflow

if TYPE_CHECKING:
    from labster.lib.workflow import State

CONVENTION = "Convention de recherche"
AVENANT_CONVENTION = "Avenant à une convention de recherche"
RECRUTEMENT = "Recrutement"
PI_LOGICIEL = "Déclaration de logiciel / base de données"
PI_INVENTION = "Déclaration d´invention"
AUTRE = "Demande autre"

TYPE_ENUM = [
    CONVENTION,
    AVENANT_CONVENTION,
    RECRUTEMENT,
    PI_LOGICIEL,
    PI_INVENTION,
    AUTRE,
]

WF_ENUM = Enum(*[s.id for s in ALL_STATES], name="demande_wf_states")

logger = structlog.get_logger()

__all__ = (
    "Demande",
    "DemandeRH",
    "DemandeConvention",
    "DemandeAvenantConvention",
    "DemandePiInvention",
    "DemandePiLogiciel",
    "DemandeAutre",
    "demande_factory",
    "CONVENTION",
    "AVENANT_CONVENTION",
    "RECRUTEMENT",
    "PI_INVENTION",
    "PI_LOGICIEL",
    "AUTRE",
)


class DemandeQuery(BaseQuery):
    def all(self) -> list[Any]:
        # print(self)
        return super().all()


class Validation:
    def __init__(
        self, obj: Demande, errors: list[Any], extra_errors: list[str]
    ) -> None:
        self.obj = obj
        self.errors = errors
        self.extra_errors = extra_errors

    @property
    def ok(self) -> bool:
        return (not self.errors) and (not self.extra_errors)


class Demande(IdMixin, TimestampedMixin, Indexable, db.Model):
    __tablename__ = "demande"
    query_class = DemandeQuery

    type = Column(Enum(*TYPE_ENUM, name="demande_type"), nullable=False, index=True)
    nom = Column(Unicode, default="", nullable=False, info=SEARCHABLE)
    name = Column(Unicode, default="", nullable=False, info=SEARCHABLE)

    no_infolab = Column(
        Unicode, default="", server_default="", nullable=False, info=SEARCHABLE
    )
    no_eotp = Column(
        Unicode, default="", server_default="", nullable=False, info=SEARCHABLE
    )

    data = Column(JSONDict(), default=dict)
    past_versions = Column(JSONList(), default=list)
    form_state = Column(JSONDict(), default=dict)
    attachments = Column(JSONDict(), default=dict)
    feuille_cout = Column(JSONDict(), default=dict)

    #: Date de validation par la hiérarchie
    date_effective = Column(Date, nullable=True)

    #: Seules les demandes actives apparaissent dans le workflow.
    #: Les autres sont considérées comme archivées.
    active = Column(Boolean, default=True, nullable=False)
    editable = Column(Boolean, default=True, nullable=False)

    # Les acteurs de la demande:

    #: id du porteur de la demande
    porteur_id = Column(Integer, ForeignKey(Profile.id), index=True)
    #: Le porteur de la demande
    porteur = relationship(Profile, foreign_keys=[porteur_id])

    #: id du gestionnaire de la demande (GDL)
    gestionnaire_id = Column(Integer, ForeignKey(Profile.id), index=True)
    #: Le gestionnaire de la demande
    gestionnaire = relationship(Profile, foreign_keys=[gestionnaire_id])

    #: id du gestionnaire de la demande (GDL)
    contact_dgrtt_id = Column(Integer, ForeignKey(Profile.id), index=True)
    #: Le gestionnaire de la demande
    contact_dgrtt = relationship(Profile, foreign_keys=[contact_dgrtt_id])

    structure_id = Column(Integer, ForeignKey(OrgUnit.id), index=True)
    structure = relationship(OrgUnit, foreign_keys=[structure_id])

    # Variables liées au workflow
    wf_state = Column(WF_ENUM, default=EN_EDITION.id, nullable=False, index=True)

    wf_date_derniere_action = Column(DateTime, nullable=False)

    #: nombre de jours de retard (0 si pas de retard)
    wf_retard = Column(Integer, nullable=False, default=0)

    wf_history = Column(JSONList(), default=list)

    wf_data = Column(JSONDict(), default=dict)

    wf_stage_id = Column(Integer, ForeignKey(OrgUnit.id), index=True, nullable=True)
    wf_stage = relationship(
        OrgUnit, primaryjoin=remote(Entity.id) == foreign(wf_stage_id)
    )

    #: id de la personne responsable de la tâche en cours
    wf_current_owner_id = Column(
        Integer, ForeignKey(Profile.id), index=True, nullable=True
    )
    #: la personne responsable de la tâche en cours
    wf_current_owner = relationship(
        Profile, primaryjoin=remote(Entity.id) == foreign(wf_current_owner_id)
    )

    __mapper_args__ = {"polymorphic_identity": "", "polymorphic_on": type}

    def __init__(self, **kw):
        super().__init__(**kw)

        assert self.porteur or self.gestionnaire
        if self.porteur:
            self.structure = self.porteur.structure

        self.data = {}
        self.attachments = {}
        self.form_state = {"fields": []}
        self.versions = []
        self.wf_state = EN_EDITION.id
        self.wf_history = []
        self.wf_data = {}
        if not self.created_at:
            self.created_at = datetime.utcnow()
        self.wf_date_derniere_action = self.created_at

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} with id={self.id}>"

    def log_creation(self, actor: Profile) -> None:
        if not self.wf_history:
            message = f"Demande créée par l'utilisateur {actor.full_name}"
            log_entry = {
                "date": datetime.now().strftime("%d %b %Y %H:%M:%S"),
                "actor_id": actor.id,
                "message": message,
                "note": "",
            }
            self.wf_history = [log_entry]

    def clone(self) -> Demande:
        nouvelle_demande = Demande(
            nom=self.nom,
            type=self.type,
            wf_state=EN_EDITION.id,
            porteur=self.porteur,
            gestionnaire=self.gestionnaire,
        )
        nouvelle_demande.data = self.data.copy()
        nouvelle_demande.form_state = self.form_state.copy()
        return nouvelle_demande

    @property
    def date_debut(self):
        return None

    @property
    def age(self):
        if not self.date_effective:
            return 0
        dt = date.today() - self.date_effective
        return int(dt.days)

    @property
    def retard(self):
        if not self.wf_date_derniere_action or not self.active:
            return 0

        # TODO: implémenter la logique de jours ouvrés
        dt = datetime.utcnow() - self.wf_date_derniere_action
        return int(dt.days)

    def update_retard(self):
        # TODO: implémenter la logique de jours ouvrés
        dt = datetime.utcnow() - self.wf_date_derniere_action
        self.wf_retard = int(dt.days)

    def nom_par_defaut(self) -> str:
        return self.type + " sans nom"

    #
    # Accessors / properties
    #
    def __getattr__(self, name: str) -> Any:
        if name.startswith("_"):
            raise AttributeError(f"object has no attribute '{name}'")

        data = getattr(self, "data", None)
        if data and name in data:
            return data[name]

        raise AttributeError(f"object has no attribute '{name}'")

    def has_same_data(self, data: dict[str, Any]) -> bool:
        """Return True if the current version is the same as the given data."""
        current_data = self.data

        for k in set(current_data.keys()) | set(data.keys()):
            if k in ["csrf_token"]:
                continue
            if k.startswith("html-"):
                continue
            current_value = current_data.get(k)
            new_value = data.get(k)
            if current_value != new_value:
                return False
        return True

    @property
    def contact(self) -> Profile:
        return self.gestionnaire or self.porteur

    @property
    def directeur_name(self) -> str:
        if self.laboratoire and self.laboratoire.directeur:
            return self.laboratoire.directeur.full_name
        return ""

    @property
    def laboratoire(self) -> OrgUnit:
        structure = self.structure
        if not structure:
            if self.porteur:
                self.structure = self.porteur.structure
            elif self.gestionnaire:
                self.structure = self.gestionnaire.laboratoire
            structure = self.structure

        return structure.laboratoire

    @property
    def owners(self) -> Collection[Profile]:
        owners = []
        if self.gestionnaire:
            owners.append(self.gestionnaire)
        if self.porteur:
            owners.append(self.porteur)
        return owners

    @property
    def contributeurs(self) -> Collection[Profile]:
        return []

    def is_editable_by(self, user: Profile):
        # FIXME
        return True
        # return self.editable and user in [self.gestionnaire, self.porteur]

    def feuille_cout_is_editable_by(self, user: Profile) -> bool:
        return self.active and user in [
            self.gestionnaire,
            self.porteur,
            self.contact_dgrtt,
        ]

    #
    # Workflow
    #
    def get_workflow(self, user: Profile | None = None) -> LabsterWorkflow:
        return LabsterWorkflow(self, user)

    def get_state(self, user: Profile | None = None) -> State:
        workflow = self.get_workflow(user)
        return workflow.current_state()

    def current_owners(self) -> list[Profile]:
        return self.get_workflow().current_owners()

    @property
    def date_soumission(self) -> date | None:
        for entry in self.wf_history:
            if entry.get("transition") == "SOUMETTRE":
                return dateutil.parser.parse(entry["date"]).date()

        return None

    @property
    def date_finalisation(self) -> date | None:
        final_states = ["CONFIRMER_FINALISATION_DGRTT", "ABANDONNER", "REJETER_DGRTT"]
        for entry in self.wf_history:
            if entry.get("transition") in final_states:
                return dateutil.parser.parse(entry["date"]).date()

        return None

    #
    # Data validation / manipulation
    #
    def validate(self) -> Validation:
        return Validation(self, self.get_errors(), self.get_extra_errors())

    def get_errors(self) -> list[Any]:
        errors = []
        form_state = self.form_state
        fields = form_state["fields"]
        for field_name, field_value in self.data.items():
            if field_name not in fields:
                continue
            field = fields[field_name]

            visible = field.get("visible")
            required = field.get("required")

            if visible and required and not field_value:
                errors.append(field["name"])
                continue

        return errors

    def get_extra_errors(self):
        return []

    def is_valid(self) -> bool:
        validation = self.validate()
        return validation.ok

    @property
    def errors(self):
        return self.validate().errors

    def update_data(self, data: dict) -> None:
        self.increase_version()
        self.data.update(data)

        self.update_nom()
        self.post_update()

    # new_name = data.get('nom', None)
    # if new_name is not None and new_name != self.name:
    #     self.name = new_name

    def post_update(self) -> None:
        new_porteur_uid = self.data.get("porteur", None)

        if new_porteur_uid is not None:
            try:
                new_porteur = Profile.query.get_by_uid(new_porteur_uid)
                self.porteur = new_porteur
                self.structure = new_porteur.structure
            except Exception:
                # TODO: better solution to deal with tests.
                pass

    def increase_version(self) -> None:
        if self.data:
            self.past_versions.append(
                (self.data, datetime.utcnow().strftime("%d %b %Y %H:%M:%S"))
            )

    # Pièces jointes
    @property
    def pieces_jointes(self) -> list[dict]:
        result = []
        for k, v in sorted(self.attachments.items()):
            if isinstance(v, dict):
                creator = Profile.query.get_by_uid(v["creator"])
                d = {"id": v["id"], "name": k, "creator": creator}

                date_str = v.get("date")
                if date_str:
                    d["date"] = iso8601.parse_date(date_str)
                else:
                    d["date"] = None
            else:
                d = {"id": v, "name": k, "creator": None, "date": None}
            id = d["id"]
            blob = Blob.query.get(id)
            if blob:
                result.append(d)

        return result


#
# Concrete classes
#
class DemandeRH(Demande):
    __mapper_args__ = {"polymorphic_identity": RECRUTEMENT}

    icon_class = "far fa-user"

    def update_nom(self) -> None:
        prenom = self.data.get("prenom") or "(prénom inconnu)"
        nom = self.data.get("nom") or "(nom inconnu)"
        nature = (
            self.data.get("nature_du_recrutement") or "(nature du recrutement inconnue)"
        )
        self.nom = f"Recrutement de {prenom} {nom} - {nature}"
        self.name = self.nom

    def assigne_contact_dgrtt(self):
        self.contact_dgrtt = get_contact_dgrtt(self.laboratoire, "CT")

    @property
    def date_debut(self) -> date | None:
        return parse_date(self.data.get("date_debut"))

    @property
    def date_fin(self) -> date | None:
        return parse_date(self.data.get("date_fin"))

    @property
    def duree_jours(self) -> int:
        if not (self.date_fin and self.date_debut):
            return 0
        return (self.date_fin - self.date_debut).days + 1

    @property
    def duree_mois(self) -> int:
        if not (self.date_fin and self.date_debut):
            return 0
        years = self.date_fin.year - self.date_debut.year
        months = self.date_fin.month - self.date_debut.month
        days = self.date_fin.day - self.date_debut.day
        return 12 * years + months + round(days / 30.4375)
        # return (self.date_fin - self.date_debut).days + 1

    @property
    def cout_total_charge(self) -> Decimal:
        try:
            return cout_total_charge(self)
        except Exception:
            return Decimal(0)

    def get_extra_errors(self) -> list[str]:
        errors: list[str] = []
        if not self.pieces_jointes:
            errors += ["Vous devez obligatoirement attacher une pièce-jointe."]
        if self.date_fin and self.date_debut and self.date_fin <= self.date_debut:
            errors += ["La date de fin est antérieure à la date de début."]
        return errors


class DemandeConvention(Demande):
    __mapper_args__ = {"polymorphic_identity": CONVENTION}

    icon_class = "far fa-briefcase"

    @property
    def date_debut(self):
        return parse_date(self.data.get("date_depot"))

    def update_nom(self):
        nom_ou_acronyme = self.data.get("nom_ou_acronyme")
        if nom_ou_acronyme:
            self.nom = nom_ou_acronyme
        else:
            self.nom = self.nom_par_defaut()

    def assigne_contact_dgrtt(self):
        """Le choix dans cette liste et la case à cocher "Le projet intègre-t-
        il une entreprise ? " pour ANR détermine le destinataire des infos du
        formulaire:

        - Si "ANR" est choisi et "non " coché = bureau des contrats publics
        - Si "ANR" est choisi et "oui " coché = bureau Entreprises
        - Si "Commission européenne" est choisi = bureau Europe
        - Si "Entreprise" est choisi = bureau Entreprises
        - Si "Financement public national" est choisi = bureau des contrats publics
        - Si "FUI" est choisi  = bureau Entreprises
        - Si "Institution européenne et internationale" est choisi = bureau des contrats publics
        - Si "Autre" est choisi  et "oui " coché = bureau Entreprises
        - Si "Autre" est choisi  et "non " coché = référent du laboratoire

        +

        Champ Type de financeur : il y a un problème d'adressage du formulaire
        quand Type de financeur=ANR.
        Si Type de financeur=ANR ET "Le projet intègre-t-il une entreprise=non"
        le formulaire est à envoyer au bureau Contrats publics
        Mais si Type de financeur=ANR ET "Le projet intègre-t-il une entreprise=oui",
        le formulaire est à envoyer au bureau Entreprises et transfert de technologie
        """
        # Rappel des bureaux
        # BureauDgrtt("ETT", "Entreprises et transfert de technologie"),
        # BureauDgrtt("CFE", "Contrats et financements européens"),
        # BureauDgrtt("CP", "Contrats publics"),
        # BureauDgrtt("CT", "Contrats de travail"),
        # BureauDgrtt("PIJ", "Propriété intellectuelle (juriste)"),
        # BureauDgrtt("PI2", "Propriété intellectuelle 2"),

        type_financeur = self.data.get("type_financeur")
        if not type_financeur:
            return

        integre_entreprise = self.data.get("integre_entreprise")
        if integre_entreprise == "non":
            integre_entreprise = False

        bureau = None
        if type_financeur == "ANR":
            if integre_entreprise:
                bureau = "ETT"
            else:
                bureau = "CP"
        elif type_financeur == "Commission Européenne":
            bureau = "CFE"
        elif type_financeur == "Entreprise":
            bureau = "ETT"
        elif type_financeur == "Financement public national":
            bureau = "CP"
        elif type_financeur == "FUI":
            bureau = "ETT"
        elif type_financeur == "Institution européenne et internationale":
            bureau = "CP"
        elif type_financeur == "Fondation":
            bureau = "CP"
        elif type_financeur == "Autre":
            if integre_entreprise:
                bureau = "ETT"
            else:
                bureau = "REF"

        assert bureau
        self.contact_dgrtt = get_contact_dgrtt(self.laboratoire, bureau)

    def get_extra_errors(self):
        if self.appel_a_projets == "oui":
            if self.pieces_jointes:
                return []
            return ["Vous devez renseigner la liste des partenaires en pièce-jointe."]

        if self.partenaires:
            return []
        return ["Vous devez saisir au moins un partenaire dans le formulaire."]


class DemandeAvenantConvention(DemandeConvention):
    __mapper_args__ = {"polymorphic_identity": AVENANT_CONVENTION}

    icon_class = "far fa-briefcase"

    def update_nom(self):
        nom_ou_acronyme = self.data.get("nom_projet")
        if nom_ou_acronyme:
            self.nom = nom_ou_acronyme
        else:
            self.nom = self.nom_par_defaut()

    def get_extra_errors(self):
        return []

    @property
    def modifications(self):
        modifs = []
        if self.montant == "oui":
            modifs.append("montant")
        if self.duree == "oui":
            modifs.append("durée")
        if self.programme_scientifique == "oui":
            modifs.append("programme scientifique")
        if self.consortium == "oui":
            modifs.append("consortium")
        if self.autre == "oui":
            modifs.append("autre")
        return ", ".join(modifs) + "."


# Abstract
class DemandePiMixin:
    intitule = ""
    acronyme = ""
    titre = ""

    def update_nom(self):
        intitule = self.intitule or self.titre
        acronyme = self.acronyme
        if acronyme and intitule:
            self.nom = f"{intitule} ({acronyme})"
        elif acronyme or intitule:
            self.nom = acronyme or intitule
        else:
            self.nom = self.nom_par_defaut()

    def assigne_contact_dgrtt(self):
        bureau = "PI2"
        self.contact_dgrtt = get_contact_dgrtt(self.laboratoire, bureau)


class DemandePiLogiciel(Demande, DemandePiMixin):
    __mapper_args__ = {"polymorphic_identity": PI_LOGICIEL}

    icon_class = "far fa-save"

    @property
    def intitule(self):
        return self.data.get("intitule", "")

    @property
    def acronyme(self):
        return self.data.get("acronyme", "")


class DemandePiInvention(Demande, DemandePiMixin):
    __mapper_args__ = {"polymorphic_identity": PI_INVENTION}

    icon_class = "far fa-rocket"

    @property
    def titre(self):
        return self.data.get("titre", "")


class DemandeAutre(Demande):
    __mapper_args__ = {"polymorphic_identity": AUTRE}

    commentaire = ""
    icon_class = "far fa-folder-open"

    @property
    def titre(self):
        return self.data.get("titre", "")

    def update_nom(self):
        self.nom = self.titre


_REGISTRY: dict[str, type[Demande]] = {
    "rh": DemandeRH,
    "convention": DemandeConvention,
    "avenant_convention": DemandeAvenantConvention,
    "pi_logiciel": DemandePiLogiciel,
    "pi_invention": DemandePiInvention,
    "autre": DemandeAutre,
}


def demande_factory(type: str, demandeur: Profile, data: dict, **args: dict) -> Demande:
    demande_cls = _REGISTRY.get(type)
    if not demande_cls:
        raise RuntimeError(f"Type de demande illégal: {type}")

    demande = demande_cls(**args)
    demande.update_data(data)
    demande.log_creation(demandeur)
    return demande



"""Methodes JSON-RPC pour manipuler le graphe des structures.

Note: les méthodes sont préfixées par "sg_" (sg = "structures graph").
"""
from __future__ import annotations

from collections import Sequence
from typing import Collection

from jsonrpcserver import method
from marshmallow import Schema, fields
from marshmallow_sqlalchemy import ModelSchema
from werkzeug.exceptions import NotFound

from labster import rbac
from labster.di import injector
from labster.domain2.model.profile import ProfileRepository
from labster.domain2.model.structure import Structure, StructureId, \
    StructureRepository
from labster.domain2.model.type_structure import ALL_TYPES, DE, EQ
from labster.domain2.services.roles import Role, RoleService
from labster.rbac import has_permission
from labster.rpc.cache import cache
from labster.security import get_current_profile, get_current_user
from labster.types import JSON
from labster.util import sort_by_name

structure_repo = injector.get(StructureRepository)
profile_repo = injector.get(ProfileRepository)
role_service = injector.get(RoleService)


@method
@cache.memoize(tag="structures", expire=3600)
def sg_all_structures() -> JSON:
    root = structure_repo.get_root()
    assert root

    def children(structure: Structure, level=0) -> list[tuple[Structure, int]]:
        assert structure

        result = [(structure, level)]
        my_children = list(structure.children)
        my_children.sort(key=lambda x: x.nom)
        for child in my_children:
            result += children(child, level + 1)
        return result

    structures2 = children(root, 0)

    entries = []
    for structure, level in structures2:
        structure_dto = {
            "id": structure.id,
            "nom": structure.nom,
            "sigle": structure.sigle,
            "type": structure.type_name,
            "level": level,
            "children_ids": [child.id for child in structure.children],
        }
        entries.append(structure_dto)

    return entries


@method
def sg_get_structure(structure_id) -> JSON:
    structure = structure_repo.get_by_id(StructureId(structure_id))
    if not structure:
        raise NotFound()

    ou_dto = FullStructureSchema().dump(structure).data
    ou_dto["parents"] = convert_structures_to_dto(list(structure.parents))
    ou_dto["children"] = convert_structures_to_dto(sort_by_name(structure.children))
    ou_dto["ancestors"] = convert_structures_to_dto(structure.ancestors)

    return ou_dto


@method
def sg_structure_can_be_deleted(id: str):
    structure = structure_repo.get_by_id(id)
    if not structure:
        raise NotFound()
    return has_permission(structure, "P3")


@method
def sg_get_possible_child_types(id: str) -> list[dict[str, str]]:
    user = get_current_profile()

    structure = structure_repo.get_by_id(id)
    if not structure:
        raise NotFound()

    if not has_permission(structure, "P3"):
        return []

    result = []
    for candidate_type in ALL_TYPES:
        if candidate_type.reel:
            continue
        if structure.type.can_have_child_of_type(candidate_type):
            if user.has_role(Role.ADMIN_CENTRAL):
                result.append({"value": candidate_type.id, "text": candidate_type.name})
            elif candidate_type in [DE, EQ]:
                result.append({"value": candidate_type.id, "text": candidate_type.name})

    return result


@method
def sg_get_parents_options(id):
    structure = structure_repo.get_by_id(id)
    if not structure:
        raise NotFound()

    if not has_permission(structure, "P2"):
        return []

    result = []
    for s in possible_parents(structure, structure_repo):
        if s.sigle:
            text = f"{s.type}: {s.name} ({s.sigle})"
        else:
            text = f"{s.type}: {s.name}"
        result.append({"text": text, "value": s.id})

    return sorted(result, key=lambda x: x["text"])


@method
def sg_get_children_options(id) -> JSON:
    structure = structure_repo.get_by_id(id)
    if not structure:
        raise NotFound()

    if not has_permission(structure, "P2"):
        return []

    result = []
    for s in possible_children(structure, structure_repo):
        if s.sigle:
            text = f"{s.type}: {s.name} ({s.sigle})"
        else:
            text = f"{s.type}: {s.name}"
        result.append({"text": text, "value": s.id})

    return sorted(result, key=lambda x: x["text"])


#
# Utils
#
def possible_parents(
    structure: Structure, repo: StructureRepository
) -> Collection[Structure]:
    s1 = repo.get_all()
    s2 = [s for s in s1 if structure.can_have_parent(s)]
    s3 = set(s2) - set(structure.parents)
    return s3


def possible_children(
    structure: Structure, repo: StructureRepository
) -> Collection[Structure]:
    s1 = repo.get_all()
    s2 = [s for s in s1 if structure.can_have_child(s)]
    s3 = set(s2) - set(structure.children)
    return s3


#
# Serialization
#
class FullStructureSchema(ModelSchema):
    is_reelle = fields.Bool()
    _depth = fields.Integer()
    can_be_deleted = fields.Method("_can_be_deleted")
    permissions = fields.Method("get_permissions")

    class Meta:
        model = Structure
        exclude = ["parents", "children"]

    def _can_be_deleted(self, structure):
        if structure.children:
            return False

        if role_service.get_users_with_given_role(Role.MEMBRE, structure):
            return False

        # For tests
        current_user = get_current_user()
        if not current_user.is_authenticated:
            return True

        profile = current_user.profile
        if profile.has_role(Role.ADMIN_CENTRAL):
            return True

        if structure.type in {DE, EQ}:
            if profile.has_role(Role.ADMIN_LOCAL, structure.parent) or profile.has_role(
                Role.ADMIN_LOCAL, structure.parent.parent
            ):

                return True

        return False

    def get_permissions(self, structure: Structure) -> dict[str, bool]:
        return {p: True for p in rbac.get_permissions_for_structure(structure)}


def convert_structures_to_dto(structures: Sequence[Structure]) -> JSON:
    class StructureSchema(Schema):
        id = fields.String()
        type = fields.String()
        nom = fields.String()
        sigle = fields.String()
        name = fields.String(attribute="sigle_ou_nom")

    return StructureSchema().dump(structures, many=True).data



""""""
from __future__ import annotations

from flask import Blueprint
from werkzeug.exceptions import Forbidden

from labster.domain2.services.roles import Role
from labster.security import get_current_profile

__all__ = ()


blueprint = Blueprint(
    "notifications", __name__, template_folder="templates", url_prefix="/notifications"
)
route = blueprint.route


@blueprint.record
def configure(state):
    from . import views


@blueprint.before_request
def check_alc_role():
    current_user = get_current_profile()
    if not current_user.has_role(Role.ADMIN_CENTRAL):
        raise Forbidden()



from __future__ import annotations

from sqlalchemy.orm import scoped_session

from labster.di import injector
from labster.domain2.model.profile import Profile, ProfileRepository
from labster.domain2.model.structure import Structure, StructureRepository
from labster.domain2.services.roles import Role, RoleService

profile_repo = injector.get(ProfileRepository)
structure_repo = injector.get(StructureRepository)
role_service = injector.get(RoleService)


def test_with_no_context(db_session: scoped_session):
    user = Profile()
    profile_repo.put(user)
    role_service.grant_role(user, Role.ADMIN_CENTRAL)
    assert role_service.has_role(user, Role.ADMIN_CENTRAL)
    assert role_service.get_users_with_role(Role.ADMIN_CENTRAL) == {user}

    role_service.clear()
    profile_repo.clear()


def test_with_context(db_session: scoped_session):
    user = Profile()
    profile_repo.put(user)

    structure = Structure()
    structure_repo.put(structure)

    role_service.grant_role(user, Role.ADMIN_LOCAL, structure)
    assert role_service.has_role(user, Role.ADMIN_LOCAL, structure)
    assert role_service.get_users_with_role(Role.ADMIN_LOCAL, structure) == {user}

    role_to_users = role_service.get_users_with_role_on(structure)
    assert len(role_to_users) == 1

    user1 = role_to_users[Role.ADMIN_LOCAL].pop()
    assert user1 == user

    roles_for_user = role_service.get_roles_for_user(user)
    assert len(roles_for_user) == 1
    assert roles_for_user[Role.ADMIN_LOCAL] == {structure}

    role_service.clear()
    profile_repo.clear()



# from __future__ import annotations
#
# from abilian.web.forms import Form, optional
# from flask import g
# from wtforms import DateField, SelectField, SelectMultipleField
#
# from labster.blueprints.bi.model import StatsLine
# from labster.blueprints.bi.util import mes_structures
# from labster.domain.models.profiles import Profile
# from labster.domain.models.unites import OrgUnit
# from labster.extensions import db
#
#
# class FilterForm(Form):
#     periode_debut = DateField(
#         "Période du", validators=(optional(),), render_kw={"placeholder": "AAAA-MM-JJ"}
#     )
#     periode_fin = DateField(
#         "Au", validators=(optional(),), render_kw={"placeholder": "AAAA-MM-JJ"}
#     )
#
#     types_demande = SelectMultipleField(
#         "Types de demande", render_kw={"class": "select2", "style": "width: 100%;"}
#     )
#     financeurs = SelectMultipleField(
#         "Financeurs", render_kw={"class": "select2", "style": "width: 100%;"}
#     )
#     types_recrutement = SelectMultipleField(
#         "Types de recrutement", render_kw={"class": "select2", "style": "width: 100%;"}
#     )
#
#     structure = SelectField(
#         "Structure",
#         validators=(optional(),),
#         render_kw={"class": "select2", "style": "width: 100%;"},
#     )
#
#     porteur = SelectField(
#         "Porteur",
#         validators=(optional(),),
#         render_kw={"class": "select2", "style": "width: 100%;"},
#     )
#
#     def __init__(self, *args, **kwargs):
#         super().__init__(*args, **kwargs)
#         self.set_type_demande_choices()
#         self.set_financeur_choices()
#         self.set_types_recrutement_choices()
#         self.set_structure_choices()
#         self.set_porteur_choices()
#
#     def set_type_demande_choices(self):
#         types_demandes = db.session.query(StatsLine.type_demande).distinct().all()
#         types_demandes.sort()
#         self.types_demande.choices = [(x[0], x[0]) for x in types_demandes]
#
#     def set_financeur_choices(self):
#         financeurs = (
#             db.session.query(StatsLine.financeur)
#             .filter(StatsLine.type_demande == "Convention de recherche")
#             .filter(StatsLine.financeur != None)
#             .filter(StatsLine.financeur != "")
#             .order_by(StatsLine.financeur)
#             .distinct()
#             .all()
#         )
#         self.financeurs.choices = [(x[0], x[0]) for x in financeurs]
#
#     def set_types_recrutement_choices(self):
#         types_recrutement = (
#             db.session.query(StatsLine.type_recrutement)
#             .filter(StatsLine.type_demande == "Recrutement")
#             .filter(StatsLine.type_recrutement != None)
#             .distinct()
#             .all()
#         )
#         self.types_recrutement.choices = [(x[0], x[0]) for x in types_recrutement]
#
#     def set_porteur_choices(self):
#         porteur_ids = (
#             db.session.query(StatsLine.porteur_id)
#             .filter(StatsLine.porteur_id != None)
#             .distinct()
#             .all()
#         )
#         porteurs = [Profile.query.get(id) for id in porteur_ids]
#         porteurs.sort(key=lambda x: (x.nom, x.prenom))
#         self.porteur.choices = [("", "-")] + [
#             (str(p.id), p.full_name) for p in porteurs
#         ]
#
#     def set_structure_choices(self):
#         equipe_ids = (
#             db.session.query(StatsLine.equipe_id)
#             .filter(StatsLine.equipe_id != None)
#             .distinct()
#             .all()
#         )
#         departement_ids = (
#             db.session.query(StatsLine.departement_id)
#             .filter(StatsLine.departement_id != None)
#             .distinct()
#             .all()
#         )
#         labo_ids = (
#             db.session.query(StatsLine.labo_id)
#             .filter(StatsLine.labo_id != None)
#             .distinct()
#             .all()
#         )
#         ufr_ids = (
#             db.session.query(StatsLine.ufr_id)
#             .filter(StatsLine.ufr_id != None)
#             .distinct()
#             .all()
#         )
#         pole_ids = (
#             db.session.query(StatsLine.pole_id)
#             .filter(StatsLine.pole_id != None)
#             .distinct()
#             .all()
#         )
#
#         user = g.current_user
#         if user.has_role("directeur"):
#             structures = mes_structures(user)
#
#         else:
#             ids = labo_ids + departement_ids + equipe_ids + ufr_ids + pole_ids
#             structure_ids = [x[0] for x in ids]
#             structures = [OrgUnit.query.get(id) for id in structure_ids]
#
#         def path(structure):
#             keys = ["pole", "ufr", "laboratoire", "departement", "equipe"]
#             path = []
#             for key in keys:
#                 value = getattr(structure, key)
#                 if value:
#                     path.append(value.nom)
#                 else:
#                     path.append("")
#             return path
#
#         to_sort = [(path(structure), structure) for structure in structures]
#         to_sort.sort()
#         structures = [t[1] for t in to_sort]
#
#         def make_label(structure):
#             prefix = "-" * structure.depth
#             return f"{prefix}{structure.nom} ({structure.type})"
#
#         self.structure.choices = [("", "-")] + [
#             (str(s.id), make_label(s)) for s in structures
#         ]
from __future__ import annotations



from __future__ import annotations

from labster.lib.workflow import Workflow

from .states import ALL_STATES, EN_EDITION
from .transitions import ABANDONNER, ACCUSER_RECEPTION, COMMENTER, \
    CONFIRMER_FINALISATION_DGRTT, CONFIRMER_RECEVABILITE_DGRTT, DESARCHIVER, \
    PRENDRE_LA_MAIN_DGRTT, PRENDRE_LA_MAIN_GESTIONNAIRE, REJETER_DGRTT, \
    REQUERIR_MODIFICATION_DGRTT, REQUERIR_MODIFICATION_DIR, SOUMETTRE, \
    VALIDER_DIR


class LabsterWorkflow(Workflow):
    initial_state = EN_EDITION

    states = ALL_STATES

    # NB: order counts!
    transitions = [
        SOUMETTRE,
        PRENDRE_LA_MAIN_GESTIONNAIRE,
        VALIDER_DIR,
        PRENDRE_LA_MAIN_DGRTT,
        REQUERIR_MODIFICATION_DIR,
        ACCUSER_RECEPTION,
        CONFIRMER_RECEVABILITE_DGRTT,
        CONFIRMER_FINALISATION_DGRTT,
        REQUERIR_MODIFICATION_DGRTT,
        REJETER_DGRTT,
        ABANDONNER,
        DESARCHIVER,
        COMMENTER,
    ]

    def actor_is_contact_labco(self):
        return self.actor == self.case.contact_labco

    def actor_is_porteur_or_gestionnaire(self):
        return self.actor in (self.case.porteur, self.case.gestionnaire)



"""API blueprint (for SPA)"""
from __future__ import annotations

# from flask import Blueprint
from flask_smorest import Blueprint

from labster.security import login_required

blueprint = Blueprint("v3", __name__, url_prefix="/v3/api")
route = blueprint.route

__all__ = ("route",)


@blueprint.before_request
@login_required
def before_request() -> None:
    # Do nothing, just require login
    pass


@blueprint.record
def configure(state):
    from . import user
